from typing import Callable, Any

import inspect


def get_available_options(_method: Callable, exclude=("self",)):
    full_args_spec = inspect.getfullargspec(_method)
    result = full_args_spec.args + full_args_spec.kwonlyargs
    return [arg for arg in result if arg not in exclude]


def safe_init(_klass, **options) -> Any:
    args = get_available_options(_klass.__init__)
    return _klass(**{key: value for key, value in options.items() if key in args})



import datetime
import jwt
from functools import partial

from sqlalchemy import select
from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint
from starlette.datastructures import MutableHeaders
from starlette.requests import HTTPConnection, Request
from starlette.types import ASGIApp, Message, Receive, Scope, Send
from starlette.responses import Response

from starlette_web.common.conf import settings
from starlette_web.common.database.session_maker import make_session_maker
from starlette_web.contrib.auth.backend import SessionJWTAuthenticationBackend
from starlette_web.contrib.auth.models import UserSession
from starlette_web.contrib.auth.utils import decode_jwt


class DBSessionMiddleware(BaseHTTPMiddleware):
    def __init__(self, app: ASGIApp) -> None:
        super().__init__(app)
        self.session_maker = make_session_maker(use_pool=False)

    async def dispatch(self, request: Request, call_next: RequestResponseEndpoint) -> Response:
        async with self.session_maker() as session:
            try:
                request.state.session = session
                request.state.db_session = session
                response = await call_next(request)
                return response
            finally:
                del request.state.session
                del request.state.db_session


class AdminSessionMiddleware:
    def __init__(
        self,
        app: ASGIApp,
    ) -> None:
        self.app = app
        self.session_cookie = SessionJWTAuthenticationBackend.cookie_name
        self.max_age = settings.AUTH_JWT_REFRESH_EXPIRES_IN
        self.path = "/"
        self.session_maker = make_session_maker(use_pool=False)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] not in ("http", "websocket"):  # pragma: no cover
            await self.app(scope, receive, send)
            return

        connection = HTTPConnection(scope)
        initial_session_was_empty = True

        if self.session_cookie in connection.cookies:
            try:
                token = connection.cookies[self.session_cookie]
                scope["session"] = decode_jwt(token)
                scope["session"]["token"] = token
                initial_session_was_empty = False
            except jwt.PyJWTError:
                scope["session"] = {}
        else:
            scope["session"] = {}

        send_wrapper = partial(
            self.send_wrapper,
            scope=scope,
            send=send,
            initial_session_was_empty=initial_session_was_empty,
        )
        await self.app(scope, receive, send_wrapper)

    async def send_wrapper(
        self,
        message: Message,
        scope: Scope,
        send: Send,
        initial_session_was_empty: bool = True,
    ) -> None:
        security_flags = "httponly; samesite=lax"
        if scope.get("scheme") == "https":
            security_flags += "; secure"

        if message["type"] == "http.response.start":
            if scope.get("session", {}) and scope["session"].get("token"):
                async with self.session_maker() as db_session:
                    query = select(UserSession).filter(
                        UserSession.refresh_token == scope["session"]["token"],
                    )
                    user_session = (await db_session.execute(query)).scalars().first()

                if user_session and not user_session.is_persistent:
                    expires = ""
                elif "exp" in scope["session"]:
                    expires = datetime.datetime.utcfromtimestamp(scope["session"]["exp"])
                    expires = expires.strftime("%a, %d %b %Y %H:%M:%S GMT")
                    expires = f"expires={expires}; "
                elif self.max_age:
                    expires = f"Max-Age={self.max_age}; "
                else:
                    expires = ""

                # We have session data to persist.
                headers = MutableHeaders(scope=message)
                header_value = "{session_cookie}={data}; path={path}; {expires}{security_flags}".format(  # noqa E501
                    session_cookie=self.session_cookie,
                    data=scope["session"]["token"],
                    path=self.path,
                    expires=expires,
                    security_flags=security_flags,
                )
                headers.append("Set-Cookie", header_value)
            elif not initial_session_was_empty:
                # The session has been cleared.
                headers = MutableHeaders(scope=message)
                header_value = "{session_cookie}={data}; path={path}; {expires}{security_flags}".format(  # noqa E501
                    session_cookie=self.session_cookie,
                    data="null",
                    path=self.path,
                    expires="expires=Thu, 01 Jan 1970 00:00:00 GMT; ",
                    security_flags=security_flags,
                )
                headers.append("Set-Cookie", header_value)

        await send(message)



from starlette.templating import Jinja2Templates

from starlette_web.common.conf import settings


templates = Jinja2Templates(
    directory=settings.TEMPLATES["ROOT_DIR"],
    autoescape=settings.TEMPLATES["AUTOESCAPE"],
    auto_reload=settings.TEMPLATES["AUTORELOAD"],
)


TemplateResponse = templates.TemplateResponse



import datetime

from starlette_web.common.management.alembic_mixin import AlembicMixin
from starlette_web.common.management.base import BaseCommand, CommandParser


class Command(BaseCommand, AlembicMixin):
    help = "Make new migrations files (helper for alembic revision)"

    def add_arguments(self, parser: CommandParser):
        parser.add_argument(
            "-n",
            "--name",
            type=str,
            required=False,
            help="New migration name",
        )

        parser.add_argument(
            "--empty",
            const=True,
            action="store_const",
            default=False,
            required=False,
            help="Whether to create an empty migration",
        )

    async def handle(self, **options):
        revision_name = "auto_" + datetime.datetime.now().strftime("%Y%m%d_%H%M")
        if options.get("name"):
            revision_name = options["name"]

        if options.get("empty", False):
            stdout, stderr = await self.run_alembic_main(["revision", "-m", revision_name])

        else:
            stdout, stderr = await self.run_alembic_main(["check"])

            if not any([
                "Target database is not up to date" in stdout,
                "No new upgrade operations detected" in stdout,
                "FAILED" in stdout,
            ]):
                stdout, stderr = await self.run_alembic_main(
                    [
                        "revision",
                        "-m",
                        revision_name,
                        "--autogenerate",
                    ]
                )

        print(stdout.strip())



import copy
import logging
import os
import tempfile

from starlette_web.contrib.scheduler.backends.base import BasePeriodicTaskScheduler
from starlette_web.contrib.scheduler.app_settings import PosixSettings


logger = logging.getLogger("starlette_web.contrib.scheduler")


class CrontabScheduler(BasePeriodicTaskScheduler):
    settings_class = PosixSettings

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.crontab_lines = []

    def _read_jobs(self):
        self.crontab_lines = os.popen(f"{self.settings.CRONTAB_EXECUTABLE} -l").readlines()

    def _write_jobs(self):
        fd, path = tempfile.mkstemp()
        tmp = os.fdopen(fd, "w")
        for line in self.crontab_lines:
            if not line.strip():
                continue
            tmp.write(line.strip() + "\n")
        tmp.close()
        os.system(f"{self.settings.CRONTAB_EXECUTABLE} {path}")
        os.unlink(path)

    def add_jobs(self):
        project_hash = self._get_project_level_hash()

        for job in self.settings.PERIODIC_JOBS:
            self.crontab_lines.append(
                self.settings.CRONTAB_LINE_PATTERN
                % {
                    "time": job[0],
                    "comment": project_hash,
                    "command": " ".join(
                        filter(
                            None,
                            [
                                self.settings.COMMAND_PREFIX,
                                self.settings.PYTHON_EXECUTABLE,
                                "command.py",
                                "scheduler",
                                "run",
                                f"--jobhash={self._hash_job(job)}",
                                "--settings=%s" % self.settings.SETTINGS_MODULE,
                                self.settings.COMMAND_SUFFIX,
                            ],
                        )
                    ),
                }
            )

    def show_jobs(self):
        project_hash = self._get_project_level_hash()

        for line in copy.copy(self.crontab_lines):
            if project_hash in line and "crontab run" in line:
                idx = line.find("crontab run")
                job_hash = line[idx:idx + 32]
                job = self._get_job_by_hash(job_hash)
                logger.info(f"{job} -> {line.strip()}")

    def remove_jobs(self):
        project_hash = self._get_project_level_hash()

        for line in copy.copy(self.crontab_lines):
            if project_hash in line:
                self.crontab_lines.remove(line)



import argparse
import os
import sys

from starlette_web.common.app import get_asgi_application
from starlette_web.common.management.base import fetch_command_by_name, CommandError


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--settings", default="core.settings", required=False)
    args, _ = parser.parse_known_args()
    os.environ.setdefault("STARLETTE_SETTINGS_MODULE", args.settings)

    if len(sys.argv) < 2:
        raise CommandError(
            'Missing command name. Correct syntax is: "python command.py command_name ..."'
        )

    from starlette_web.common.conf import settings

    command = fetch_command_by_name(sys.argv[1])
    app = get_asgi_application(use_pool=settings.DB_USE_CONNECTION_POOL_FOR_MANAGEMENT_COMMANDS)
    command(app).run_from_command_line(sys.argv)



# flake8: noqa

from starlette_web.contrib.camel_case.parser import CamelCaseStarletteParser
from starlette_web.contrib.camel_case.renderer import CamelCaseJSONRenderer
from starlette_web.contrib.camel_case.utils import camelize, underscoreize



from typing import Type

from marshmallow import Schema, fields

from starlette_web.common.conf import settings
from starlette_web.common.http.exceptions import ImproperlyConfigured
from starlette_web.common.utils import import_string


__ERROR_RESPONSE_SCHEMA = None


class ErrorResponseSchema(Schema):
    error = fields.String(required=True, allow_none=False)
    details = fields.Raw(required=False, allow_none=True)


def get_error_schema_class() -> Type[Schema]:
    global __ERROR_RESPONSE_SCHEMA
    if __ERROR_RESPONSE_SCHEMA is not None:
        return __ERROR_RESPONSE_SCHEMA

    try:
        __ERROR_RESPONSE_SCHEMA = import_string(settings.ERROR_RESPONSE_SCHEMA)
    except (SystemError, ImportError, TypeError, ValueError):
        raise ImproperlyConfigured(
            details=(
                "Invalid value settings.ERROR_RESPONSE_SCHEMA "
                f"= {settings.ERROR_RESPONSE_SCHEMA}"
            )
        )

    return __ERROR_RESPONSE_SCHEMA



from typing import Any

from starlette_web.common.channels.event import Event
from starlette_web.common.http.exceptions import NotSupportedError


class BaseChannelLayer:
    def __init__(self, **options) -> None:
        pass

    async def connect(self) -> None:
        raise NotSupportedError()

    async def disconnect(self) -> None:
        raise NotSupportedError()

    async def subscribe(self, group: str, **kwargs) -> None:
        raise NotSupportedError()

    async def unsubscribe(self, group: str, **kwargs) -> None:
        raise NotSupportedError()

    async def publish(self, group: str, message: Any, **kwargs) -> None:
        raise NotSupportedError()

    async def next_published(self) -> Event:
        raise NotSupportedError()



import json
import pickle
from typing import Any

from starlette_web.common.http.exceptions import BaseApplicationError


class SerializerError(BaseApplicationError):
    pass


class SerializeError(SerializerError):
    pass


class DeserializeError(SerializerError):
    pass


class BaseSerializer:
    def serialize(self, content: Any) -> Any:
        raise NotImplementedError

    def deserialize(self, content: Any) -> Any:
        raise NotImplementedError

    def __eq__(self, other):
        return self.__class__ == other.__class__

    @staticmethod
    def serializes_to_bytes():
        return False


class BytesSerializer(BaseSerializer):
    @staticmethod
    def serializes_to_bytes():
        return True


class JSONSerializer(BaseSerializer):
    encoder_class = json.JSONEncoder
    decoder_class = json.JSONDecoder

    def serialize(self, content: Any) -> Any:
        try:
            return self.encoder_class().encode(content)
        except ValueError as exc:
            raise SerializeError from exc

    def deserialize(self, content: Any) -> Any:
        try:
            return self.decoder_class().decode(content or "null")
        except ValueError as exc:
            raise DeserializeError from exc


class PickleSerializer(BytesSerializer):
    # TODO: maybe add option to use dill ?
    def serialize(self, content: Any) -> Any:
        try:
            return pickle.dumps(content)
        except pickle.PicklingError as exc:
            raise SerializeError from exc

    def deserialize(self, content: Any) -> Any:
        if content is None:
            return None

        try:
            return pickle.loads(content)
        except pickle.UnpicklingError as exc:
            raise DeserializeError from exc



from typing import Any, Type

from redis import asyncio as aioredis
from redis.asyncio.client import PubSub
from redis.exceptions import ConnectionError

from starlette_web.common.channels.event import Event
from starlette_web.common.channels.exceptions import ListenerClosed
from starlette_web.common.channels.layers.base import BaseChannelLayer
from starlette_web.common.utils.encoding import force_str
from starlette_web.common.utils.serializers import BytesSerializer, PickleSerializer


class RedisPubSubChannelLayer(BaseChannelLayer):
    # Cross-process channel layer, uses fire-and-forget scheme
    # If you are using sharded redis, you'll have to inherit the class
    # and redefine some commands to their sharded versions
    serializer_class: Type[BytesSerializer] = PickleSerializer
    redis: aioredis.Redis

    def __init__(self, **options):
        super().__init__(**options)
        self.redis = aioredis.Redis(**options)
        self._serializer = self.serializer_class()
        self._pubsub: PubSub = self.redis.pubsub()

    async def connect(self) -> None:
        await self._pubsub.connect()

    async def disconnect(self) -> None:
        await self._pubsub.aclose()

    async def subscribe(self, group: str, **kwargs) -> None:
        await self._pubsub.subscribe(group)

    async def unsubscribe(self, group: str, **kwargs) -> None:
        await self._pubsub.unsubscribe(group)

    async def publish(self, group: str, message: Any, **kwargs) -> None:
        message = self._serializer.serialize(message)
        await self.redis.publish(group, message)

    async def next_published(self) -> Event:
        while True:
            try:
                response = await self._pubsub.parse_response(block=True)
            except ConnectionError as exc:
                raise ListenerClosed(details=str(exc)) from exc

            if response is None:
                continue

            message = await self._pubsub.handle_message(response, ignore_subscribe_messages=True)
            if message is None:
                continue

            return Event(
                group=force_str(message["channel"]),
                message=self._serializer.deserialize(message["data"]),
            )



import logging
import logging.config
from typing import List, Union, Dict, Callable, Type, TypeVar, Optional

from sqlalchemy.orm import sessionmaker
from starlette.applications import Starlette
from starlette.middleware import Middleware
from starlette.requests import Request
from starlette.routing import Route, Mount, WebSocketRoute
from starlette.types import Lifespan
from webargs_starlette import WebargsHTTPException

from starlette_web.common.caches import caches
from starlette_web.common.conf import settings
from starlette_web.common.conf.app_manager import app_manager
from starlette_web.common.database import make_session_maker
from starlette_web.common.http.exception_handlers import (
    BaseApplicationErrorHandler,
    WebargsHTTPExceptionHandler,
)
from starlette_web.common.http.renderers import BaseRenderer
from starlette_web.common.http.exceptions import BaseApplicationError
from starlette_web.common.utils import import_string, get_available_options


AppClass = TypeVar("AppClass", bound=Starlette)
ExceptionHandlerType = Callable[[Request, Exception], BaseRenderer]


class WebApp(Starlette):
    """Simple adaptation of Starlette APP. Small addons here."""

    session_maker: sessionmaker

    def __init__(self, *args, **kwargs):
        use_pool = kwargs.pop("use_pool", True)

        allowed_params = get_available_options(Starlette.__init__)
        starlette_init_kwargs = {
            key: value for key, value in kwargs.items() if key in allowed_params
        }
        super().__init__(*args, **starlette_init_kwargs)

        # TODO: multiple database support (planned to 0.4)
        if settings.DATABASE_DSN:
            self.session_maker = make_session_maker(use_pool=use_pool)
        else:

            def _not_implemented():
                raise NotImplementedError("settings.DATABASE_DSN is not configured")

            self.session_maker = _not_implemented


class BaseStarletteApplication:
    app_class: AppClass = WebApp

    def __init__(self, **kwargs):
        self.run_checks = kwargs.get("run_checks", True)
        self._event_handlers = []

    def pre_app_init(self) -> None:
        """
        Extra actions before app's initialization (can be overridden)
        """
        app_manager.initialize_apps()
        if self.run_checks:
            app_manager.run_apps_checks()

    def post_app_init(self, app: AppClass) -> None:
        """
        Extra actions after app's initialization (can be overridden)
        """
        pass

    def get_debug(self) -> bool:
        return settings.APP_DEBUG

    def get_middlewares(self) -> List[Middleware]:
        return settings.MIDDLEWARES

    def get_routes(self) -> List[Union[WebSocketRoute, Route, Mount]]:
        routes_setting = settings.ROUTES
        if routes_setting is None:
            return []

        return import_string(routes_setting)

    def get_exception_handlers(self) -> Dict[Type[Exception], ExceptionHandlerType]:
        return {
            BaseApplicationError: BaseApplicationErrorHandler(),
            WebargsHTTPException: WebargsHTTPExceptionHandler(),
        }

    def get_lifespan(self) -> Optional[Lifespan["AppClass"]]:
        """
        This method returns None or AsyncContextManager,
        that is instantiated with application instance
        and yields either None or State (dict, typeddict)
        See example of state here https://www.starlette.io/lifespan/

        If you need to open multiple async context managers within lifespan context,
        consider using contextlib.AsyncExitStack
        # flake8: noqa
        # https://anyio.readthedocs.io/en/stable/cancellation.html#avoiding-cancel-scope-stack-corruption

        >>> class AppLifespan:
        >>>     def __init__(self, app: WebApp):
        >>>         self.app = app
        >>>
        >>>     async def __aenter__(self):
        >>>         print("START")
        >>>         return {}
        >>>
        >>>     async def __aexit__(self, exc_type, exc_val, exc_tb):
        >>>         print("END")
        >>>         return False
        >>>
        >>> return AppLifespan
        """

        return None

    def get_app(self, **kwargs) -> AppClass:
        object.__getattribute__(settings, "_setup")()
        self.pre_app_init()

        app = self.app_class(
            routes=self.get_routes(),
            exception_handlers=self.get_exception_handlers(),
            debug=self.get_debug(),
            middleware=self.get_middlewares(),
            lifespan=self.get_lifespan(),
            **kwargs,
        )

        self._setup_logging(app)
        self._setup_caches(app)

        self.post_app_init(app)
        return app

    def _setup_logging(self, app: AppClass):
        if settings.LOGGING:
            logging.config.dictConfig(settings.LOGGING)

    def _setup_caches(self, app: AppClass):
        for conn_name in settings.CACHES:
            _ = caches[conn_name]


def get_asgi_application(**kwargs) -> AppClass:
    StarletteApplication = import_string(settings.APPLICATION_CLASS)
    run_checks = kwargs.pop("run_checks_on_startup", True)
    return StarletteApplication(run_checks=run_checks).get_app(**kwargs)



import json
from typing import Any

from starlette.responses import Response, BackgroundTask

from starlette_web.common.utils import StarletteJSONEncoder


class BaseRenderer(Response):
    pass


class JSONRenderer(BaseRenderer):
    json_encode_class = StarletteJSONEncoder
    media_type = "application/json"

    # Copied from starlette.responses.JSONResponse
    def __init__(
        self,
        content: Any,
        status_code: int = 200,
        headers: dict = None,
        media_type: str = None,
        background: BackgroundTask = None,
    ) -> None:
        super().__init__(content, status_code, headers, media_type, background)

    def render(self, content: Any) -> bytes:
        return json.dumps(
            self.preprocess_content(content),
            ensure_ascii=False,
            allow_nan=False,
            indent=None,
            separators=(",", ":"),
            cls=self.json_encode_class,
        ).encode("utf-8")

    @staticmethod
    def preprocess_content(content):
        return content



from starlette.requests import Request
from starlette.types import Scope

from starlette_web.contrib.auth.models import User
from starlette_web.common.authorization.permissions import BasePermission
from starlette_web.common.http.exceptions import PermissionDeniedError


class IsSuperuserPermission(BasePermission):
    async def has_permission(self, request: Request, scope: Scope) -> bool:
        user: User = scope.get("user")
        if (not user) or (not getattr(user, "is_superuser", False)):
            raise PermissionDeniedError(details="You don't have an admin privileges.")

        return True



from typing import Dict, Any, List

from sqlalchemy import select
from sqlalchemy.orm import sessionmaker
from sqlalchemy.exc import IntegrityError

from starlette_web.common.database import make_session_maker
from starlette_web.contrib.constance.backends.base import BaseConstanceBackend
from starlette_web.contrib.constance.backends.database.models import Constance


class DatabaseBackend(BaseConstanceBackend):
    session_maker: sessionmaker

    def __init__(self):
        super().__init__()
        # Disable connection pooling to avoid creating persistent open connections
        self.session_maker = make_session_maker(use_pool=False)

    async def mget(self, keys: List[str]) -> Dict[str, Any]:
        async with self.session_maker() as session:
            query = select(Constance)

            constants = (await session.execute(query)).scalars()
            values = {key: self.empty for key in keys}
            values = {
                **values,
                **{
                    constant.key: self._preprocess_response(constant.value)
                    for constant in constants
                    if constant.key in keys
                },
            }
            return values

    async def get(self, key: str) -> Any:
        async with self.session_maker() as session:
            query = select(Constance).filter(Constance.key == key)
            val = (await session.execute(query)).scalars().first()
            return self._preprocess_response(val.value if val else None)

    async def set(self, key: str, value: Any) -> None:
        async with self.session_maker() as session:
            query = select(Constance).filter(Constance.key == key).with_for_update()
            instance = (await session.execute(query)).scalars().first()

            if instance is None:
                try:
                    instance = Constance(
                        key=key,
                        value=self.serializer.serialize(value),
                    )
                    session.add(instance)
                    await session.commit()
                except IntegrityError:
                    # This case may occur, if multiple processes try
                    # to update an existing value at the same time
                    # TODO: examine, whether this could be processes in a better way
                    pass

            else:
                instance.value = self.serializer.serialize(value)
                await session.commit()



from sqlalchemy import select

from starlette_web.common.management.base import BaseCommand, CommandError
from starlette_web.contrib.auth.models import User
from starlette_web.contrib.auth.management.auth_command_mixin import AuthCommandMixin


class Command(AuthCommandMixin, BaseCommand):
    help = "Change a password of existing User (contrib.auth)"

    async def handle(self, **options):
        async with self.app.session_maker() as session:
            email = self.get_input_data("Input email (username): ")
            self.validate_field("email", email)

            query = select(User).filter(User.email == email)
            user = (await session.execute(query)).scalars().first()
            if user is None:
                raise CommandError(details=f"User with email = {email} not found.")

            password_1 = self.get_input_data("Input new password: ")
            self.validate_field("password", password_1)
            password_2 = self.get_input_data("Retype new password: ")
            if password_1 != password_2:
                raise CommandError(details="Password mismatch.")

            self.validate_password(password_1, user=user)

            user.password = User.make_password(password_1)
            await session.commit()

            print(f"Password for user {user} updated successfully.")



from difflib import SequenceMatcher
from functools import cache
from typing import Optional

from starlette_web.common.authorization.base_user import BaseUserMixin
from starlette_web.common.conf import settings
from starlette_web.common.http.exceptions import NotSupportedError, InvalidParameterError
from starlette_web.common.utils.importing import import_string


# TODO: maybe raise marshmallow.ValidationError ?
# TODO: add common password validators
class BasePasswordValidator:
    def __init__(self, **options):
        self.options = options

    def __call__(self, password: str, user: Optional[BaseUserMixin] = None):
        raise NotSupportedError()


class PasswordLengthValidator(BasePasswordValidator):
    def __init__(self, **options):
        super().__init__(**options)
        self.min_length = options.get("MIN_LENGTH", 8)
        if type(self.min_length) is not int or self.min_length < 0:
            raise NotSupportedError(details="Invalid input options for MIN_LENGTH.")
        self.max_length = options.get("MAX_LENGTH", 64)
        if type(self.max_length) is not int or self.max_length < 0:
            raise NotSupportedError(details="Invalid input options for MAX_LENGTH.")

    def __call__(self, password: str, user: Optional[BaseUserMixin] = None):
        if len(password) < self.min_length:
            raise InvalidParameterError(details=f"Password length is less than {self.min_length}.")

        if len(password) > self.max_length:
            raise InvalidParameterError(details=f"Password length is more than {self.max_length}.")


class UsernameSimilarityValidator(BasePasswordValidator):
    def __init__(self, **options):
        super().__init__(**options)
        self.max_similarity = options.get("MAX_SIMILARITY", 0.7)
        if type(self.max_similarity) is not float or not (0 <= self.max_similarity <= 1):
            raise NotSupportedError(details="Invalid input options for MAX_SIMILARITY.")

    def __call__(self, password: str, user: Optional[BaseUserMixin] = None):
        if not user:
            return

        if not hasattr(user, "username_field") or not user.username_field:
            return

        username_field = user.username_field

        try:
            username = str(getattr(user, username_field))
        except:  # noqa:
            return

        if (
            SequenceMatcher(
                a=password.lower(),
                b=username.lower(),
            ).quick_ratio()
            > self.max_similarity
        ):
            raise InvalidParameterError(details="Password is too similar with username.")


class NumericPasswordValidator(BasePasswordValidator):
    def __call__(self, password: str, user: Optional[BaseUserMixin] = None):
        if password.isdigit():
            raise InvalidParameterError(details="Password is entirely numeric.")


@cache
def _get_validators():
    validators = []
    for validator_setting in settings.AUTH_PASSWORD_VALIDATORS:
        validator_class = import_string(validator_setting["BACKEND"])
        options = validator_setting.get("OPTIONS", {})
        validators.append(validator_class(**options))
    return validators


def validate_password(password, user: Optional[BaseUserMixin] = None):
    for validator in _get_validators():
        validator(password, user)



from importlib import import_module
import os
from typing import Dict, Any

from starlette_web.common.http.exceptions import ImproperlyConfigured


_getattr = object.__getattribute__
_setattr = object.__setattr__
empty = object()


class Settings:
    ENVIRONMENT_VARIABLE = "STARLETTE_SETTINGS_MODULE"
    _global_settings: Dict[str, Any]
    _user_settings: Dict[str, Any]
    _setup_done: bool

    def __init__(self):
        _setattr(self, "_global_settings", dict())
        _setattr(self, "_user_settings", dict())
        _setattr(self, "_setup_done", False)

    def _setup(self) -> None:
        settings_module = os.environ.get(_getattr(self, "ENVIRONMENT_VARIABLE"))
        if not settings_module:
            raise ImproperlyConfigured(
                details="Environment variable STARLETTE_SETTINGS_MODULE is not configured."
            )

        module = import_module(settings_module)
        for setting in dir(module):
            if setting.isupper():
                setting_value = getattr(module, setting)
                _getattr(self, "_user_settings")[setting] = setting_value

        global_settings_module = import_module("starlette_web.common.conf.global_settings")
        for setting in dir(global_settings_module):
            if setting.isupper():
                setting_value = getattr(global_settings_module, setting)
                _getattr(self, "_global_settings")[setting] = setting_value

        _getattr(self, "_user_settings")["STARLETTE_SETTINGS_MODULE"] = settings_module
        _setattr(self, "_setup_done", True)

    def __getattr__(self, key: str) -> Any:
        if not _getattr(self, "_setup_done"):
            _getattr(self, "_setup")()

        value = _getattr(self, "_user_settings").get(key, empty)

        if value is empty:
            value = _getattr(self, "_global_settings").get(key, empty)

            if value is empty:
                raise ImproperlyConfigured(details=f"Setting {key.upper()} is not configured.")

        return value

    def __setattr__(self, key: str, value: Any) -> None:
        _getattr(self, "_user_settings")[key.upper()] = value


settings = Settings()



import re


def redis_pattern_to_re_pattern(pattern: str) -> re.Pattern:
    # flake8: noqa
    # Source: https://github.com/jamesls/fakeredis/pull/184/files#diff-abc1f5739597cb0cca80af31f32bc494b8d994472058fafdc57dba7ca4a8732dR242
    """
    A helper function to match redis key pattern with re built-in module,
    Redis key pattern has different semantics, than regex.
    Official docs for redis key matching https://redis.io/commands/keys/

    >>> import re
    >>> from starlette_web.common.utils.regex import redis_pattern_to_re_pattern

    >>> redis_pattern = "keys:*"
    >>> re_pattern = redis_pattern_to_re_pattern(redis_pattern)
    >>> re_pattern
    [0] re.compile('keys:.*')
    >>> re.fullmatch(re_pattern, "keys:1")
    [1] <re.Match object; span=(0, 6), match='keys:1'>
    >>> re.fullmatch(re_pattern, "key:1")
    [2] None
    """
    parts = ["^"]
    i = 0
    L = len(pattern)
    while i < L:
        c = pattern[i]
        if c == "?":
            parts.append(".")
        elif c == "*":
            parts.append(".*")
        elif c == "\\":
            if i < L - 1:
                i += 1
            parts.append(re.escape(pattern[i]))
        elif c == "[":
            parts.append("[")
            i += 1
            if i < L and pattern[i] == "^":
                i += 1
                parts.append("^")
            while i < L:
                if pattern[i] == "\\":
                    i += 1
                    if i < L:
                        parts.append(re.escape(pattern[i]))
                elif pattern[i] == "]":
                    break
                elif i + 2 <= L and pattern[i + 1] == "-":
                    start = pattern[i]
                    end = pattern[i + 2]
                    if start > end:
                        start, end = end, start
                    parts.append(re.escape(start) + "-" + re.escape(end))
                    i += 2
                else:
                    parts.append(re.escape(pattern[i]))
                i += 1
            parts.append("]")
        else:
            parts.append(re.escape(pattern[i]))
        i += 1
    parts.append("\\Z")
    regex = "".join(parts)
    return re.compile(regex, re.S)



import hashlib
import os
import pickle
import tempfile
import time
from typing import Any, Union, Optional

import anyio
from anyio.lowlevel import cancel_shielded_checkpoint
from filelock import FileLock as StrictFileLock

from starlette_web.common.conf import settings
from starlette_web.common.caches.base_lock import BaseLock, CacheLockError


class FileLock(BaseLock):
    """
    An async variation of SoftFileLock with support of timeout (via os.path.getmtime)
    """

    EXIT_MAX_DELAY = 60.0

    def __init__(
        self,
        name: Union[str, os.PathLike[Any]],
        timeout: Optional[float] = None,
        blocking_timeout: Optional[float] = None,
        **kwargs,
    ) -> None:
        super().__init__(name, timeout, blocking_timeout, **kwargs)
        self._stored_file_ts = {}
        self._retry_interval = kwargs.get("retry_interval", 0.001)

    async def _acquire(self):
        if self._is_acquired:
            return

        while True:
            await anyio.sleep(self._retry_interval)
            try:
                with self._get_manager_lock():
                    if self._sync_acquire():
                        self._acquire_event.set()
                        return
            except CacheLockError:
                self._task_group.cancel_scope.cancel()
                raise
            except OSError:
                continue

    async def _release(self):
        if not self._is_acquired:
            return

        while True:
            try:
                await cancel_shielded_checkpoint()
            except BaseException:  # noqa
                self._sync_release()
                raise

            try:
                with self._get_manager_lock():
                    self._sync_release()
                    return
            except OSError:
                continue

    def _sync_release(self):
        try:
            ts = os.path.getmtime(self._name)
            # Another process has re-acquired lock due to timeout
            if ts not in self._stored_file_ts:
                return
            os.unlink(self._name)
        finally:
            self._stored_file_ts = {}
            self._is_acquired = False

    def _sync_acquire(self) -> bool:
        if os.path.exists(self._name):
            ts = os.path.getmtime(self._name)

            if ts not in self._stored_file_ts:
                with open(self._name, "rb") as file:
                    try:
                        self._stored_file_ts[ts] = pickle.loads(file.read())
                    except pickle.PickleError as exc:
                        raise CacheLockError(details=str(exc)) from exc

            # Timeout on other instance has not expired
            if self._stored_file_ts[ts] + ts > time.time():
                return False

            # Timeout for lock has expired, so we acquire it.
            # Remove lock file to update its mtime
            self._sync_release()

        with open(self._name, "wb+") as file:
            file.write(pickle.dumps(self._timeout, protocol=4))
            # Guarantee that writing is finalized
            file.flush()
            os.fsync(file.fileno())

        ts = os.path.getmtime(self._name)
        self._stored_file_ts[ts] = self._timeout
        return True

    def __del__(self):
        try:
            self._sync_release()
        except OSError:
            pass

    def _get_manager_lock(self):
        _manager_lock_filepath = os.path.join(
            tempfile.gettempdir(),
            self._get_project_hash() + "_filelock.lock",
        )
        # timeout=0 means exactly 1 attempt to acquire lock
        return StrictFileLock(_manager_lock_filepath, timeout=0)

    def _get_project_hash(self):
        return hashlib.md5(str(settings.SECRET_KEY).encode("utf-8")).hexdigest()



# flake8: noqa

from starlette.routing import Mount, Route, WebSocketRoute
from starlette.staticfiles import StaticFiles

from starlette_web.common.conf import settings
from starlette_web.contrib.apispec.views import OpenApiView
from starlette_web.contrib.admin import admin, AdminMount


routes = [
    Mount(
        settings.STATIC["URL"],
        app=StaticFiles(directory=settings.STATIC["ROOT_DIR"]),
        name="static",
    ),
    Mount(
        settings.MEDIA["URL"],
        app=StaticFiles(directory=settings.MEDIA["ROOT_DIR"]),
        name="media",
    ),
    # Admin-panel is not yet integral part of the framework,
    # so it requires a bit of hacking with custom mount class,
    # to manage static files right
    AdminMount("/admin", app=admin.get_app(), name="admin"),
    Route("/openapi", OpenApiView, include_in_schema=False),
]



import json
import pickle
from typing import Any

from starlette_web.common.utils.serializers import (
    BytesSerializer,
    SerializeError,
    DeserializeError,
)


class MQTTSerializer(BytesSerializer):
    def serialize(self, content: Any) -> Any:
        if isinstance(content, (bytes, bytearray, memoryview)):
            return bytes(content)

        if isinstance(content, str):
            return content.encode()

        try:
            return json.dumps(content).encode()
        except (TypeError, ValueError):
            pass

        try:
            return pickle.dumps(content)
        except pickle.PickleError as exc:
            raise SerializeError from exc

    def deserialize(self, content: Any) -> Any:
        try:
            return json.loads(content)
        except (TypeError, ValueError, json.JSONDecodeError):
            pass

        try:
            return pickle.loads(content)
        except pickle.UnpicklingError:
            pass

        try:
            return content.decode()
        except UnicodeDecodeError as exc:
            raise DeserializeError from exc



import anyio
from anyio.streams.memory import (
    MemoryObjectReceiveStream,
    MemoryObjectSendStream,
)
from typing import Dict, Any, Optional

from starlette_web.common.channels.event import Event
from starlette_web.common.channels.layers.base import BaseChannelLayer


class InMemoryChannelLayer(BaseChannelLayer):
    # Single-process channel layer, uses fire-and-forget scheme
    def __init__(self, max_buffer_size: int = 0, **options):
        super().__init__(**options)
        self._subscribed: Dict = dict()
        self._receive_stream: Optional[MemoryObjectReceiveStream] = None
        self._send_stream: Optional[MemoryObjectSendStream] = None
        self.max_buffer_size = max_buffer_size
        self._manager_lock = anyio.Lock()

    async def connect(self) -> None:
        self._send_stream, self._receive_stream = anyio.create_memory_object_stream(
            max_buffer_size=self.max_buffer_size,
        )

    async def disconnect(self) -> None:
        self._subscribed.clear()
        self._send_stream.close()
        self._receive_stream.close()
        self._send_stream = None
        self._receive_stream = None

    async def subscribe(self, group: str, **kwargs) -> None:
        async with self._manager_lock:
            self._subscribed.setdefault(group, 0)
            self._subscribed[group] += 1

    async def unsubscribe(self, group: str, **kwargs) -> None:
        async with self._manager_lock:
            self._subscribed.setdefault(group, 0)
            self._subscribed[group] -= 1

    async def publish(self, group: str, message: Any, **kwargs) -> None:
        if not self._send_stream:
            raise RuntimeError(".publish() requires not-null self._send_stream")

        event = Event(group=group, message=message)
        await self._send_stream.send(event)

    async def next_published(self) -> Event:
        if not self._receive_stream:
            raise RuntimeError(".next_published() requires not-null self._receive_stream")

        while True:
            event = await self._receive_stream.receive()
            if self._subscribed.get(event.group, 0) > 0:
                return event



from typing import Type, Any, Optional, Dict, Sequence, AsyncContextManager, List

from starlette_web.common.http.exceptions import BaseApplicationError
from starlette_web.common.utils import Singleton
from starlette_web.common.utils.serializers import BaseSerializer, PickleSerializer


class CacheError(BaseApplicationError):
    message = "Cache error."


class CacheLockError(CacheError):
    message = "Failed to lock or unlock a cache."


class BaseCache(metaclass=Singleton):
    serializer_class: Type[BaseSerializer] = PickleSerializer
    serializer: BaseSerializer

    def __init__(self, options: Dict[str, Any]):
        self.serializer = self.serializer_class()

    async def async_get(self, key: str) -> Any:
        raise NotImplementedError

    async def async_set(self, key: str, value: Any, timeout: Optional[float] = 120) -> None:
        raise NotImplementedError

    async def async_delete(self, key: str) -> None:
        raise NotImplementedError

    async def async_keys(self, pattern: str) -> List[str]:
        # Returns list of keys, matching redis-like pattern
        # See docs: https://redis.io/commands/keys/
        raise NotImplementedError

    async def async_has_key(self, key: str) -> bool:
        raise NotImplementedError

    async def async_get_many(self, keys: Sequence[str]) -> Dict[str, Any]:
        result = dict()
        for key in keys:
            result[key] = await self.async_get(key)
        return result

    async def async_set_many(self, data: Dict[str, Any], timeout: Optional[float] = 120) -> None:
        for key, value in data.items():
            await self.async_set(key, value, timeout=timeout)

    async def async_delete_many(self, keys: Sequence[str]) -> None:
        for key in keys:
            await self.async_delete(key)

    async def async_clear(self) -> None:
        raise NotImplementedError

    def lock(
        self,
        name: str,
        timeout: Optional[float] = 20,
        blocking_timeout: Optional[float] = None,
        **kwargs,
    ) -> AsyncContextManager:
        raise NotImplementedError



# flake8: noqa

from starlette_web.common.files.storages.filesystem import (
    FilesystemStorage,
    MediaFileSystemStorage,
)
from starlette_web.common.files.storages.storage_manager import storage_manager



import logging
import logging.config
from typing import Any, Optional

import httpx
from starlette.requests import Request
from starlette.responses import BackgroundTask
from webargs_starlette import WebargsHTTPException

from starlette_web.common.conf import settings
from starlette_web.common.http.exceptions import (
    BaseApplicationError,
    InvalidParameterError,
)
from starlette_web.common.http.renderers import BaseRenderer, JSONRenderer
from starlette_web.common.http.schemas import get_error_schema_class


class BaseExceptionHandler:
    renderer_class: BaseRenderer = JSONRenderer

    def _log_message(self, exc: Exception, error_data: dict, level=logging.ERROR):
        logger = logging.getLogger(__name__)

        error_details = {
            "error": error_data.get("error", "Unbound exception"),
            "details": error_data.get("details", str(exc)),
        }
        message = "{exc.__class__.__name__} '{error}': [{details}]".format(exc=exc, **error_details)
        logger.log(level, message, exc_info=(level == logging.ERROR))

    def _get_error_message(self, request: Request, exc: Exception) -> str:
        return "Something went wrong!"

    def _get_error_details(self, request: Request, exc: Exception) -> str:
        return f"Raised Error: {exc.__class__.__name__}"

    def _get_status_code(self, request: Request, exc: Exception) -> int:
        return getattr(exc, "status_code", BaseApplicationError.message)

    def _get_response_data(self, request: Request, exc: Exception) -> Any:
        _status_code = self._get_status_code(request, exc)

        payload = {
            "error": self._get_error_message(request, exc),
        }
        if any([
            settings.ERROR_DETAIL_FORCE_SUPPLY,
            settings.APP_DEBUG,
            _status_code == InvalidParameterError.status_code,
        ]):
            payload["details"] = self._get_error_details(request, exc)

        error_schema = get_error_schema_class()()
        return error_schema.dump(payload)

    def _on_error_action(self, request: Request, exc: Exception):
        status_code = self._get_status_code(request, exc)
        error_message = self._get_error_message(request, exc)
        payload = {"error": error_message}

        log_level = logging.ERROR if httpx.codes.is_error(status_code) else logging.WARNING
        self._log_message(exc, payload, log_level)

    def _get_headers(self, request: Request, exc: Exception) -> Optional[dict]:
        return None

    def _get_background_tasks(self, request: Request, exc: Exception) -> Optional[BackgroundTask]:
        return None

    def __call__(self, request: Request, exc: Exception) -> BaseRenderer:
        self._on_error_action(request, exc)

        return self.renderer_class(
            content=self._get_response_data(request, exc),
            status_code=self._get_status_code(request, exc),
            background=self._get_background_tasks(request, exc),
            headers=self._get_headers(request, exc),
        )


class BaseApplicationErrorHandler(BaseExceptionHandler):
    def _get_error_details(self, request: Request, exc: BaseApplicationError) -> str:
        return exc.details

    def _get_error_message(self, request: Request, exc: BaseApplicationError) -> str:
        return exc.message


class WebargsHTTPExceptionHandler(BaseExceptionHandler):
    def _get_error_details(self, request: Request, exc: WebargsHTTPException):
        return exc.messages.get("json") or exc.messages.get("form") or exc.messages

    def _get_error_message(self, request: Request, exc: WebargsHTTPException) -> str:
        return InvalidParameterError.message

    def _get_status_code(self, request: Request, exc: Exception) -> int:
        return InvalidParameterError.status_code



from apispec import APISpec
from starlette.requests import Request

from starlette_web.common.conf import settings
from starlette_web.common.http.base_endpoint import BaseHTTPEndpoint
from starlette_web.common.http.exceptions import NotImplementedByServerError
from starlette_web.common.http.renderers import JSONRenderer
from starlette_web.common.http.responses import TemplateResponse
from starlette_web.common.utils import urljoin
from starlette_web.contrib.apispec.introspection import APISpecSchemaGenerator
from starlette_web.contrib.apispec.marshmallow import StarletteWebMarshmallowPlugin


api_spec = APISpec(
    **settings.APISPEC["CONFIG"],
    plugins=[StarletteWebMarshmallowPlugin()],
)


schemas = APISpecSchemaGenerator(api_spec)


# TODO: think about managing permissions
# TODO: cache view
class OpenApiView(BaseHTTPEndpoint):
    auth_backend = None
    permission_classes = []

    async def get(self, request: Request):
        _format = request.query_params.get("format", "openapi")
        _schema_url = request.url.replace_query_params(format="openapi")

        if _format not in ["openapi", "redoc"]:
            raise NotImplementedByServerError(
                details=(
                    "Server only supports 'redoc' and 'openapi' "
                    "format for OpenAPI documentation."
                )
            )

        if _format == "openapi":
            routes = request.app.routes
            return JSONRenderer(schemas.get_schema(routes))

        return TemplateResponse(
            request=request,
            name="apispec/redoc.html",
            context={
                "REDOC_SPEC_URL": str(_schema_url),
                "REDOC_JS_URL": urljoin(settings.STATIC["URL"], "apispec", "redoc.js"),
                "REDOC_TITLE": "OPENAPI documentation",
                "request": request,
            },
        )



from typing import Any, List, Dict

from starlette_web.common.conf import settings
from starlette_web.common.http.exceptions import (
    NotSupportedError,
    BaseApplicationError,
    UnexpectedError,
    ImproperlyConfigured,
)
from starlette_web.common.utils import import_string
from starlette_web.contrib.constance.backends.base import BaseConstanceBackend
from starlette_web.contrib.constance.backends.caching_mixin import ConstanceCacheMixin


# TODO: some kind of validation system, much like django.checks
# TODO: _validate method
class LazyConstance:
    _backend: BaseConstanceBackend
    _is_setup: bool

    def __init__(self):
        self._is_setup = False

    async def get(self, key: str) -> Any:
        self._setup()

        if key not in settings.CONSTANCE_CONFIG:
            raise NotSupportedError(
                details=f"Key {key} is not defined in settings.CONSTANCE_CONFIG."
            )

        value = await self._backend.get(key)
        return self._postprocess_value(key, value)

    async def set(self, key: str, value: Any) -> None:
        self._setup()

        if key not in settings.CONSTANCE_CONFIG:
            raise NotSupportedError(
                details=f"Key {key} is not defined in settings.CONSTANCE_CONFIG."
            )

        expected_type = settings.CONSTANCE_CONFIG[key][2]
        if type(value) is not expected_type:
            try:
                # This is for cases like int-float, str-int and such
                value = expected_type(value)
            except (TypeError, ValueError) as exc:
                raise NotSupportedError(details=str(exc))
            except Exception as exc:
                # I.e. if we try to pass datetime to uuid value, which causes OverflowError
                raise UnexpectedError(details=str(exc))

        await self._backend.set(key, value)

    async def mget(self, keys: List[str]) -> Dict[str, Any]:
        self._setup()

        if not keys:
            return {}

        for key in keys:
            if key not in settings.CONSTANCE_CONFIG:
                raise NotSupportedError(
                    details=f"Key {key} not present in settings.CONSTANCE_CONFIG."
                )

        return_list = await self._backend.mget(keys)
        return {key: self._postprocess_value(key, value) for key, value in return_list.items()}

    def _postprocess_value(self, key, value):
        if key not in settings.CONSTANCE_CONFIG:
            raise NotSupportedError(details=f"Key {key} not present in settings.CONSTANCE_CONFIG.")

        if value == self._backend.empty:
            return settings.CONSTANCE_CONFIG[key][0]

        return value

    def _setup(self) -> None:
        if self._is_setup:
            return

        try:
            _import_path = settings.CONSTANCE_BACKEND
            if _import_path is None:
                _backend_kls = None
            else:
                _backend_kls = import_string(_import_path)
        except (AttributeError, BaseApplicationError):
            _backend_kls = None
        except (ImportError, SystemError):
            raise ImproperlyConfigured(
                details=f"Invalid constance class: {settings.CONSTANCE_BACKEND}"
            )

        try:
            _cache_key = settings.CONSTANCE_DATABASE_CACHE_BACKEND
        except (AttributeError, BaseApplicationError):
            _cache_key = None

        if _backend_kls:
            if _cache_key:
                self._backend = type(
                    _backend_kls.__name__,
                    (ConstanceCacheMixin, _backend_kls),
                    {"_cache_key": _cache_key},
                )()
            else:
                self._backend = _backend_kls()

        self._is_setup = True


config = LazyConstance()



# Copied from https://github.com/django/django/blob/main/django/core/serializers/json.py

import json
import decimal
import datetime
import uuid


def _get_duration_components(duration):
    days = duration.days
    seconds = duration.seconds
    microseconds = duration.microseconds

    minutes = seconds // 60
    seconds = seconds % 60

    hours = minutes // 60
    minutes = minutes % 60

    return days, hours, minutes, seconds, microseconds


def _duration_iso_string(duration):
    if duration < datetime.timedelta(0):
        sign = "-"
        duration *= -1
    else:
        sign = ""

    days, hours, minutes, seconds, microseconds = _get_duration_components(duration)
    ms = ".{:06d}".format(microseconds) if microseconds else ""
    return "{}P{}DT{:02d}H{:02d}M{:02d}{}S".format(sign, days, hours, minutes, seconds, ms)


class StarletteJSONEncoder(json.JSONEncoder):
    """
    JSONEncoder subclass that knows how to encode date/time, decimal types, and UUIDs.

    >>> import json
    >>> import datetime
    >>> from decimal import Decimal
    >>> from starlette_web.common.utils import StarletteJSONEncoder
    >>> obj = {'key_decimal': Decimal('10.02'), 'key_timedelta': datetime.timedelta(days=1, seconds=13876)}  ## noqa E501
    >>> obj = {**obj, 'key_time': datetime.time(13,3,6,999), "key_date": datetime.date(2020,1,1)}
    >>> json.dumps(obj, cls=StarletteJSONEncoder)
    [0] '{"key_decimal": "10.02", "key_timedelta": "P1DT03H51M16S", "key_time": "13:03:06.000", "key_date": "2020-01-01"}'  ## noqa E501
    """

    def default(self, o):
        # See "Date Time String Format" in the ECMA-262 specification.
        if isinstance(o, datetime.datetime):
            r = o.isoformat()
            if o.microsecond:
                r = r[:23] + r[26:]
            if r.endswith("+00:00"):
                r = r[:-6] + "Z"
            return r
        elif isinstance(o, datetime.date):
            return o.isoformat()
        elif isinstance(o, datetime.time):
            return o.isoformat()
        elif isinstance(o, datetime.timedelta):
            return _duration_iso_string(o)
        elif isinstance(o, (decimal.Decimal, uuid.UUID)):
            return str(o)
        else:
            return super().default(o)



from referencing.exceptions import PointerToNowhere
from traceback_with_variables import format_exc

from starlette_web.common.conf import settings
from starlette_web.common.conf.base_app_config import BaseAppConfig
from starlette_web.common.http.exceptions import ImproperlyConfigured
from starlette_web.common.utils import import_string
from starlette_web.contrib.apispec.views import schemas


class AppConfig(BaseAppConfig):
    app_name = "apispec"

    def initialize(self):
        try:
            __import__("openapi_spec_validator")
        except (SystemError, ImportError):
            raise ImproperlyConfigured(
                details=(
                    "Extra dependency 'openapi_spec_validator' is required"
                    " for starlette_web.contrib.apispec "
                    "Install it via 'pip install starlette-web[apispec]'."
                )
            )

    def perform_checks(self):
        from openapi_spec_validator import validate
        from openapi_spec_validator.validation.exceptions import (
            OpenAPIValidationError,
            OpenAPISpecValidatorError,
        )

        routes = import_string(settings.ROUTES)

        try:
            # This check mostly fails on invalid indentation
            # or partially missing properties.
            api_spec = schemas.get_schema(routes)
        except Exception as exc:  # noqa
            # Printing variable values in traceback is the only
            # viable way to know which schema is fallible
            raise ImproperlyConfigured(
                message="Invalid schema in apispec",
                details=format_exc(exc, num_skipped_frames=0),
            )

        try:
            # This check finds missing whole blocks,
            # i.e. missing info about path parameter in schema
            validate(api_spec)
        except (OpenAPIValidationError, OpenAPISpecValidatorError) as exc:
            raise ImproperlyConfigured(details=str(exc))
        except PointerToNowhere as exc:
            _str_exc = str(exc).split(" ")
            if _str_exc[1:4] == ['does', 'not', 'exist']:
                _name = _str_exc[0]
                if _name.startswith("'/components/schemas/") and _name.endswith("'"):
                    _name = _name[21:-1]
                    from marshmallow.class_registry import get_class
                    _klass_list = get_class(_name, all=True)
                    if type(_klass_list) is list and len(_klass_list) > 1:
                        pass
                    else:
                        _klass_list = []
                else:
                    _klass_list = []

                error_description = (
                    f"Schema component {_str_exc[0]} cannot be found in OpenAPI schema. "
                    f"This is most likely caused by creating multiple subclasses of "
                    f"marshmallow.schema.Schema with same class name. "
                    f"Ensure that all classes have unique names."
                )

                if _klass_list:
                    error_description += "\nDetected schemas with same class name:\n" + "\n".join([
                        f"{kls.__module__}.{kls.__name__}" for kls in _klass_list
                    ])

                raise ImproperlyConfigured(details=error_description) from exc
            else:
                raise ImproperlyConfigured(details=str(exc)) from exc



from starlette_web.common.conf.base_app_config import BaseAppConfig
from starlette_web.contrib.auth.password_validation import _get_validators


class AppConfig(BaseAppConfig):
    app_name = "auth"

    def initialize(self):
        _get_validators()



# Modification of https://github.com/Woile/starlette-apispec/blob/master/starlette_apispec/schemas.py  # noqa: E501
from collections import deque
from typing import List

from apispec import APISpec
from apispec.exceptions import DuplicateComponentNameError
from starlette.routing import BaseRoute, Mount, Route
from starlette.schemas import BaseSchemaGenerator, EndpointInfo

from starlette_web.common.authorization.backends import BaseAuthenticationBackend
from starlette_web.common.http.exceptions import ImproperlyConfigured
from starlette_web.common.http.schemas import get_error_schema_class
from starlette_web.common.i18n import gettext


# TODO: custom EndpointInfo with link to endpoint class
# TODO: maybe allow user to override class in settings (?)
# TODO: allow user to define rules to format path parameters with custom converters
class APISpecSchemaGenerator(BaseSchemaGenerator):
    ERROR_SCHEMA_NAME = "Error"

    def __init__(self, spec: APISpec) -> None:
        self.spec: APISpec = spec
        self.security_schemas = {}
        self.paths_and_endpoints = {}
        self._original_paths = {}

    def _fetch_paths_and_endpoints(self, initial_routes: List[BaseRoute]):
        if self.paths_and_endpoints:
            return self.paths_and_endpoints

        mounts = deque([])
        routes = {}

        for route in initial_routes:
            _cleaned_path = self._remove_converter(route.path)
            # Skip any websocket connections
            if isinstance(route, Mount):
                mounts.append((_cleaned_path, route))
                self._original_paths[_cleaned_path] = route.path
            elif isinstance(route, Route) and getattr(route, "include_in_schema", True):
                routes[self._remove_converter(route.path)] = route.endpoint
                self._original_paths[_cleaned_path] = route.path

        while mounts:
            path, mount = mounts.popleft()
            _original_path = self._original_paths[path]
            for route in mount.routes:
                _cleaned_path = self._remove_converter(route.path)
                # Skip any websocket connections
                if isinstance(route, Mount):
                    mounts.append((path + _cleaned_path, route))
                    self._original_paths[path + _cleaned_path] = _original_path + route.path
                elif isinstance(route, Route) and getattr(route, "include_in_schema", True):
                    routes[path + _cleaned_path] = route.endpoint
                    self._original_paths[path + _cleaned_path] = _original_path + route.path
            self._original_paths.pop(path, None)

        self.paths_and_endpoints = routes

    def _add_security_schema(self, auth_backend: BaseAuthenticationBackend) -> str:
        auth_name = getattr(auth_backend, "openapi_name", None) or auth_backend.__name__

        if auth_name in self.security_schemas:
            if self.security_schemas[auth_name] != auth_backend:
                raise DuplicateComponentNameError(
                    gettext("Duplicate auth_backend $auth_name", auth_name=auth_name)
                )
        else:
            self.security_schemas[auth_name] = auth_backend
            self.spec.components.security_scheme(auth_name, auth_backend.openapi_spec)

        return auth_name

    def _populate_security_schema(self, endpoint: EndpointInfo, parsed_docstring: dict):
        if parsed_docstring.get("security"):
            return

        kls = self.paths_and_endpoints[endpoint.path]
        if not getattr(kls, "auth_backend", None):
            return

        if getattr(kls.auth_backend, "openapi_spec", None):
            security_schema = self._add_security_schema(kls.auth_backend)
            # TODO: examine, if something should be passed into dict object
            parsed_docstring["security"] = [{security_schema: []}]

    def _populate_auth_errors(self, endpoint: EndpointInfo, parsed_docstring: dict):
        kls = self.paths_and_endpoints[endpoint.path]

        if getattr(kls, "auth_backend", None):
            parsed_docstring["responses"] = {
                **parsed_docstring.get("responses", {}),
                "401": {
                    "description": gettext("Authentication error."),
                    "content": {"application/json": {"schema": self.ERROR_SCHEMA_NAME}},
                },
            }

        if getattr(kls, "permission_classes", []):
            parsed_docstring["responses"] = {
                **parsed_docstring.get("responses", {}),
                "403": {
                    "description": gettext("Access forbidden."),
                    "content": {"application/json": {"schema": self.ERROR_SCHEMA_NAME}},
                },
            }

    def _populate_validation(self, endpoint: EndpointInfo, parsed_docstring: dict):
        if "requestBody" in parsed_docstring:
            parsed_docstring["responses"] = {
                **parsed_docstring.get("responses", {}),
                "400": {
                    "description": gettext("Bad Request."),
                    "content": {"application/json": {"schema": self.ERROR_SCHEMA_NAME}},
                },
            }

    def _populate_path_params(self, endpoint: EndpointInfo, parsed_docstring: dict):
        try:
            original_path = self._original_paths[endpoint.path]
        except KeyError:
            return

        _parameters_to_add = {}

        start_pos = original_path.find("{")
        while start_pos > -1:
            end_pos = original_path.find("}")
            if end_pos < start_pos + 2:
                break

            param = original_path[start_pos + 1:end_pos]

            if ":" in param:
                if param.count(":") > 1:
                    raise ImproperlyConfigured(
                        details=f"Invalid definition of path parameter in endpoint "
                        f"{self._original_paths[endpoint.path]}. "
                        f"Only single converter is allowed.",
                    )

                param_name, param_converter = param.split(":")
            else:
                param_name, param_converter = param, "str"

            if param_name == "" or param_converter == "":
                break

            _parameters_to_add[param_name] = param_converter
            original_path = original_path[end_pos + 1:]
            start_pos = original_path.find("{")

        if _parameters_to_add:
            parsed_docstring.setdefault("parameters", [])
            parameters = parsed_docstring["parameters"]
            for parameter in parameters:
                if parameter["in"] == "path":
                    _parameters_to_add.pop(parameter["name"], None)

            for param_name, param_converter in _parameters_to_add.items():
                schema = {"type": "string"}
                if param_converter == "int":
                    schema = {"type": "integer"}
                elif param_converter == "float":
                    schema = {"type": "number", "format": "double"}
                elif param_converter in ["uuid", "path"]:
                    schema["format"] = param_converter
                parsed_docstring["parameters"].append({
                    "in": "path",
                    "required": True,
                    "name": param_name,
                    "schema": schema,
                })

    def get_schema(self, routes: List[BaseRoute]) -> dict:
        ErrorResponseSchema = get_error_schema_class()()
        if self.ERROR_SCHEMA_NAME not in self.spec.components.schemas:
            self.spec.components.schema(self.ERROR_SCHEMA_NAME, schema=ErrorResponseSchema)

        endpoints = self.get_endpoints(routes)
        self._fetch_paths_and_endpoints(routes)

        for endpoint in endpoints:
            parsed = self.parse_docstring(endpoint.func)
            self._populate_security_schema(endpoint, parsed)
            self._populate_auth_errors(endpoint, parsed)
            self._populate_validation(endpoint, parsed)
            self._populate_path_params(endpoint, parsed)

            self.spec.path(
                path=endpoint.path,
                operations={endpoint.http_method: parsed},
            )

        return self.spec.to_dict()



import os
import pkgutil
from argparse import ArgumentParser
from difflib import get_close_matches
from functools import partial
from typing import Optional, List, Type, Dict, Coroutine, Any, Callable

import anyio
from anyio._core._eventloop import T_Retval

from starlette_web.common.app import get_asgi_application, WebApp
from starlette_web.common.conf import settings
from starlette_web.common.http.exceptions import BaseApplicationError
from starlette_web.common.utils import import_string


class CommandError(BaseApplicationError):
    message = "Error while running a user-defined command."


class CommandParser(ArgumentParser):
    """
    Customized ArgumentParser class to improve some error messages and prevent
    SystemExit in several occasions, as SystemExit is unacceptable when a
    command is called programmatically.
    """

    def __init__(self, *, missing_args_message=None, called_from_command_line=None, **kwargs):
        self.called_from_command_line = called_from_command_line
        super().__init__(**kwargs)

    def error(self, message):
        if self.called_from_command_line:
            super().error(message)
        else:
            raise CommandError(message)


class BaseCommand:
    help: Optional[str] = None

    def __init__(self, app):
        self.parser: Optional[CommandParser] = None
        self.app: WebApp = app

    def create_parser(self, argv, called_from_command_line=True):
        parser = CommandParser(
            prog="%s %s" % (argv[0], argv[1]),
            description=self.help or None,
            called_from_command_line=called_from_command_line,
        )

        return parser

    def add_arguments(self, parser: CommandParser):
        # Redefine in inherited classes
        pass

    async def handle(self, **options):
        raise NotImplementedError

    async def _handle_wrapper(self, **options):
        async with self.app.router.lifespan_context(self.app) as state:
            await self.handle(_lifespan_state=state, **options)

    def prepare_command_function(
        self,
        argv,
        called_from_command_line,
    ) -> Callable[..., Coroutine[Any, Any, T_Retval]]:
        self.parser = self.create_parser(
            argv,
            called_from_command_line=called_from_command_line,
        )
        self.add_arguments(self.parser)
        namespace, _ = self.parser.parse_known_args(args=argv[2:])
        kwargs = namespace.__dict__
        return partial(self._handle_wrapper, **kwargs)

    def run_from_command_line(self, argv: List[str]):
        func = self.prepare_command_function(argv, True)
        anyio.run(func)

    async def run_from_code(self, argv: List[str]):
        coroutine = self.prepare_command_function(argv, False)()
        await coroutine


def _get_app_path_by_name(app: str) -> str:
    """
    Lookup possible locations for management commands.
    Firstly, local project directories are checked.
    If respective module was not found,
    trying to seek in virtual environment.
    """
    module_name, _, module_rest = app.partition(".")
    found_local_module = None
    for module_info in pkgutil.iter_modules([module_name]):
        if module_info.ispkg:
            found_local_module = module_info.name

    app_path = app.replace(".", os.sep)

    if not found_local_module:
        try:
            venv_module = __import__(module_name)
            prefix = os.path.dirname(venv_module.__path__[0]).rstrip(os.sep)
        except (OSError, SystemError, ImportError, IndexError):
            prefix = ""

        if prefix:
            app_path = prefix.rstrip(os.sep) + os.sep + app_path.lstrip(os.sep)

    return app_path


def list_commands() -> Dict[str, str]:
    command_files = {}

    installed_apps = settings.INSTALLED_APPS
    if "starlette_web.common" not in installed_apps:
        installed_apps = ["starlette_web.common"] + installed_apps

    for app in installed_apps:
        modules = [os.sep.join([_get_app_path_by_name(app), "management", "commands"])]
        for module_info in pkgutil.iter_modules(modules):
            if module_info.name.startswith("_") or module_info.ispkg:
                continue

            if module_info.name in command_files:
                raise CommandError(
                    details=f"Command '{module_info.name}' is declared in multiple modules."
                )

            command_files[module_info.name] = ".".join(
                [app, "management", "commands", module_info.name, "Command"]
            )

    return command_files


def fetch_command_by_name(command_name: str) -> Type[BaseCommand]:
    commands = list_commands()

    if command_name in commands:
        try:
            command = import_string(commands[command_name])

            # isinstance not working for class, imported with import_module
            is_instance = any([kls for kls in command.__mro__ if kls == BaseCommand])
            if not is_instance:
                raise CommandError(
                    "Command must be inherited from common.management.base.BaseCommand"
                )

            return command
        except (ImportError, AssertionError) as exc:
            raise CommandError from exc

    error_message = f"Command '{command_name}' not found."
    possible_matches = get_close_matches(command_name, list(commands.keys()))
    if possible_matches:
        details = f'Did you mean "{possible_matches[0]}"?'
    else:
        details = ""
    raise CommandError(message=error_message, details=details)


async def call_command(command_name, command_args: List[str]):
    command = fetch_command_by_name(command_name)
    app = get_asgi_application(
        use_pool=settings.DB_USE_CONNECTION_POOL_FOR_MANAGEMENT_COMMANDS,
        run_checks_on_startup=False,
    )
    await command(app).run_from_code(["command.py", command_name] + command_args)



import os
import sys


def main():
    current_settings = os.environ.get("STARLETTE_SETTINGS_MODULE")

    try:
        # At this point, user-defined settings may not exist
        # (i.e., if user calls startproject command),
        # so pass global_settings instead
        settings_module = "starlette_web.common.conf.global_settings"

        sys_argv = list(sys.argv).copy()
        for arg in sys_argv.copy():
            if arg.startswith("--settings="):
                settings_module = arg[11:]
                sys_argv.remove(arg)

        os.environ.setdefault("STARLETTE_SETTINGS_MODULE", settings_module)

        if len(sys_argv) < 2:
            raise Exception(
                "Missing command name. Correct syntax is: " '"starlette-web-admin command_name ..."'
            )

        from starlette_web.common.conf import settings
        from starlette_web.common.app import get_asgi_application
        from starlette_web.common.management.base import fetch_command_by_name

        command = fetch_command_by_name(sys_argv[1])
        app = get_asgi_application(use_pool=settings.DB_USE_CONNECTION_POOL_FOR_MANAGEMENT_COMMANDS)
        command(app).run_from_command_line(sys_argv)
    finally:
        if current_settings:
            os.environ["STARLETTE_SETTINGS_MODULE"] = current_settings
        else:
            os.unsetenv("STARLETTE_SETTINGS_MODULE")


if __name__ == "__main__":
    sys.path.append(os.getcwd())
    main()



# Core application settings

APP_DEBUG = False
APPLICATION_CLASS = "starlette_web.common.app.BaseStarletteApplication"

SECRET_KEY = ""
INSTALLED_APPS = []
MIDDLEWARES = []
ROUTES = None

LOGGING = {}

# Used in contrib.auth
SITE_URL = "http://localhost:80"

# TODO: for future support of i18n
LANGUAGE_CODE = "en-us"

# Database settings

DB_ASYNC_SESSION_CLASS = "sqlalchemy.ext.asyncio.AsyncSession"
DB_USE_CONNECTION_POOL_FOR_MANAGEMENT_COMMANDS = False
DB_POOL_RECYCLE = 3600
DB_ECHO = False

# Must be overridden in user-defined settings
DATABASE_DSN = None

# Common.cache

CACHES = {
    "default": {
        "BACKEND": "starlette_web.common.caches.local_memory.LocalMemoryCache",
        "OPTIONS": {
            "name": "default",
        },
    },
}

# Common.http

ERROR_RESPONSE_SCHEMA = "starlette_web.common.http.schemas.ErrorResponseSchema"
DEFAULT_REQUEST_PARSER = "webargs_starlette.StarletteParser"
DEFAULT_RESPONSE_RENDERER = "starlette_web.common.http.renderers.JSONRenderer"
STATUS_CODES_WITH_NO_BODY = {100, 101, 102, 103, 204, 304}
REMOVE_BODY_FROM_RESPONSE_WITH_NO_BODY = False
ERROR_DETAIL_FORCE_SUPPLY = False

# Common.files

STORAGES = {
    "default": {
        "BACKEND": "starlette_web.common.files.storages.MediaFileSystemStorage",
    },
}

# Common.email

EMAIL_SENDER = None

# Contrib.auth

AUTH_PASSWORD_HASHERS = [
    "starlette_web.contrib.auth.hashers.PBKDF2PasswordHasher",
]

AUTH_PASSWORD_VALIDATORS = [
    {"BACKEND": "starlette_web.contrib.auth.password_validation.NumericPasswordValidator"},
    {"BACKEND": "starlette_web.contrib.auth.password_validation.PasswordLengthValidator"},
    {"BACKEND": "starlette_web.contrib.auth.password_validation.UsernameSimilarityValidator"},
]

# see https://pyjwt.readthedocs.io/en/latest/algorithms.html for details
AUTH_JWT_EXPIRES_IN = 300  # 5 min
AUTH_JWT_REFRESH_EXPIRES_IN = 30 * 24 * 3600  # 30 days
AUTH_JWT_ALGORITHM = "HS512"
AUTH_INVITE_LINK_EXPIRES_IN = 3 * 24 * 3600  # 3 day
AUTH_RESET_PASSWORD_LINK_EXPIRES_IN = 3 * 3600  # 3 hours

# Contrib.constance

CONSTANCE_CONFIG = {}
CONSTANCE_DATABASE_CACHE_BACKEND = None
CONSTANCE_BACKEND = None

# Contrib.scheduler

PERIODIC_JOBS_BACKEND = None



import logging
from typing import ClassVar, Type, Any, Optional, List, Dict, Tuple
import sys

import anyio
from anyio._core._tasks import TaskGroup
from marshmallow import Schema, ValidationError
from starlette.endpoints import WebSocketEndpoint
from starlette.types import Scope, Receive, Send
from starlette.websockets import WebSocket, WebSocketDisconnect

from starlette_web.common.app import WebApp
from starlette_web.common.authorization.backends import (
    BaseAuthenticationBackend,
    NoAuthenticationBackend,
)
from starlette_web.common.authorization.base_user import BaseUserMixin, AnonymousUser
from starlette_web.common.authorization.permissions import PermissionType
from starlette_web.common.http.exceptions import (
    PermissionDeniedError,
    AuthenticationFailedError,
)
from starlette_web.common.utils.crypto import get_random_string


logger = logging.getLogger(__name__)


class BaseWSEndpoint(WebSocketEndpoint):
    auth_backend: ClassVar[Type[BaseAuthenticationBackend]] = NoAuthenticationBackend
    permission_classes: ClassVar[List[PermissionType]] = []
    request_schema: ClassVar[Type[Schema]]
    user: BaseUserMixin
    task_group: Optional[TaskGroup]
    EXIT_MAX_DELAY: float = 60
    encoding = "json"

    def __init__(self, scope: Scope, receive: Receive, send: Send) -> None:
        super().__init__(scope, receive, send)
        self.task_group = None
        self.app: WebApp = self.scope.get("app")

    async def dispatch(self) -> None:
        async with anyio.create_task_group() as self.task_group:
            await super().dispatch()

    async def on_connect(self, websocket: WebSocket) -> None:
        try:
            async with self.app.session_maker() as db_session:
                websocket.state.db_session = db_session
                self.user = await self._authenticate(websocket)
                permitted, reason = await self._check_permissions(websocket)
        except Exception as exc:  # pylint: disable=broad-except
            # Any exception must be re-raised to WebSocketDisconnect
            # Otherwise, websocket will hang out,
            # since accept()/close() have not been called
            raise WebSocketDisconnect(code=1006, reason=str(exc)) from exc

        # Explicitly clear db_session,
        # so that user does not use it through lengthy websocket life-state
        del websocket.state.db_session

        if permitted:
            await websocket.accept()
        else:
            raise WebSocketDisconnect(code=3000, reason=reason)

    async def on_receive(self, websocket: WebSocket, data: Any) -> None:
        cleaned_data = self._validate(data)
        task_id = get_random_string(50)
        self.task_group.start_soon(self._background_handler_wrap, task_id, websocket, cleaned_data)

    async def on_disconnect(self, websocket: WebSocket, close_code: int) -> None:
        if self.task_group:
            if sys.exc_info() == (None, None, None):
                self.task_group.cancel_scope.cancel()
                logger.debug("TaskGroup has been explicitly cancelled.")
            else:
                logger.debug("TaskGroup will be implicitly cancelled due to exception.")

        logger.debug("WS connection has been closed.")

    async def _background_handler_wrap(self, task_id: str, websocket: WebSocket, data: Dict):
        task_result = None

        with anyio.CancelScope() as cancel_scope:
            await self._register_background_task(task_id, websocket, data)

            try:
                try:
                    task_result = await self._background_handler(task_id, websocket, data)
                except anyio.get_cancelled_exc_class() as exc:
                    logger.debug(f"Background task {task_id} has been cancelled.")
                    # As per anyio documentation, CancelError must be always re-raised.
                    # Note, that it will also cancel all other tasks.
                    raise exc
                except Exception as exc:  # pylint: disable=broad-except
                    await self._handle_background_task_exception(task_id, websocket, exc)
                finally:
                    cancel_scope.deadline = anyio.current_time() + self.EXIT_MAX_DELAY
                    cancel_scope.shield = True
            finally:
                await self._unregister_background_task(task_id, websocket, task_result)

    async def _handle_background_task_exception(
        self, task_id: str, websocket: WebSocket, exc: Exception
    ):
        error_message = "Couldn't finish _background_handler for class %s"
        error_message_message_args = (self.__class__.__name__,)
        logger.exception(error_message, *error_message_message_args)

        # By default, an exception is propagated to the parent TaskGroup,
        # causing an inner CancelScope to cancel all child tasks.
        # If this is not an expected behavior, you may silence this error.
        raise WebSocketDisconnect(code=1005, reason=str(exc)) from exc

    async def _register_background_task(self, task_id: str, websocket: WebSocket, data: Dict):
        # This method is to be redefined in child classes
        pass

    async def _unregister_background_task(
        self, task_id: str, websocket: WebSocket, task_result: Any
    ):
        # This method is to be redefined in child classes
        pass

    async def _background_handler(self, task_id: str, websocket: WebSocket, data: Dict) -> Any:
        raise WebSocketDisconnect(
            code=1005,
            reason="Background handler for Websocket is not implemented",
        )

    def _validate(self, request_data: dict) -> dict:
        try:
            return self.request_schema().load(request_data)
        except ValidationError as exc:
            # TODO: check that details is str / flatten
            raise WebSocketDisconnect(
                code=1007,
                reason=exc.data,
            ) from exc

    async def _authenticate(self, websocket: WebSocket) -> BaseUserMixin:
        backend = self.auth_backend(websocket, self.scope)
        try:
            user = await backend.authenticate()
        except AuthenticationFailedError:
            user = AnonymousUser()
        self.scope["user"] = user
        return user

    async def _check_permissions(self, websocket: WebSocket) -> Tuple[bool, str]:
        for permission_class in self.permission_classes:
            try:
                has_permission = await permission_class().has_permission(websocket, self.scope)
                if not has_permission:
                    return False, PermissionDeniedError.message
            except PermissionDeniedError as exc:
                return False, str(exc)

        return True, ""



from sqlalchemy import LargeBinary, Column, String

from starlette_web.common.database import ModelBase


class Constance(ModelBase):
    __tablename__ = "constance"

    key = Column(String(length=255), primary_key=True)
    value = Column(LargeBinary(), nullable=True, default="")

    def __str__(self):
        return self.key



import logging
from typing import Tuple

from sqlalchemy import select
from jwt import InvalidTokenError, ExpiredSignatureError

from starlette_web.contrib.auth.models import User, UserSession
from starlette_web.contrib.auth.utils import decode_jwt, TOKEN_TYPE_ACCESS
from starlette_web.common.authorization.backends import BaseAuthenticationBackend
from starlette_web.common.http.exceptions import (
    AuthenticationFailedError,
    AuthenticationRequiredError,
    SignatureExpiredError,
)

logger = logging.getLogger(__name__)


class JWTAuthenticationBackend(BaseAuthenticationBackend):
    """Core of authenticate system, based on JWT auth approach"""

    keyword = "Bearer"
    openapi_spec = {"type": "http", "scheme": "bearer", "bearerFormat": "JWT"}
    openapi_name = "JWTAuth"

    async def authenticate(self, **kwargs) -> User:
        request = self.request

        auth_header = request.headers.get("Authorization") or request.headers.get("authorization")
        if not auth_header:
            raise AuthenticationRequiredError("Invalid token header. No credentials provided.")

        auth = auth_header.split()
        if len(auth) != 2:
            logger.warning("Trying to authenticate with header %s", auth_header)
            raise AuthenticationFailedError("Invalid token header. Token should be format as JWT.")

        if auth[0] != self.keyword:
            raise AuthenticationFailedError("Invalid token header. Keyword mismatch.")

        user, _, session_id = await self.authenticate_user(jwt_token=auth[1], **kwargs)

        self.scope["user_session_id"] = session_id
        self.scope["user"] = user
        return user

    @staticmethod
    def _parse_jwt_payload(jwt_token: str, token_type: str) -> dict:
        logger.debug("Logging via JWT auth. Got token: %s", jwt_token)
        try:
            # TODO: class-based JWT decoder
            jwt_payload = decode_jwt(jwt_token)
        except ExpiredSignatureError:
            logger.debug("JWT signature has been expired for token %s", jwt_token)
            raise SignatureExpiredError("JWT signature has been expired for token")
        except InvalidTokenError as error:
            msg = "Token could not be decoded: %s"
            logger.exception(msg, error)
            raise AuthenticationFailedError(msg % (error,))

        if jwt_payload["token_type"] != token_type:
            raise AuthenticationFailedError(
                f"Token type '{token_type}' expected, got '{jwt_payload['token_type']}' instead."
            )

        return jwt_payload

    async def authenticate_user(
        self,
        jwt_token: str,
        token_type: str = TOKEN_TYPE_ACCESS,
        **kwargs,
    ) -> Tuple[User, dict, str]:
        """Allows to find active user by jwt_token"""
        jwt_payload = self._parse_jwt_payload(jwt_token, token_type)

        user_id = jwt_payload.get("user_id")

        query = select(User).filter(User.id == user_id, User.is_active.is_(True))
        user = (await self.request.state.db_session.execute(query)).scalars().first()

        if not user:
            msg = "Couldn't found active user with id=%s."
            logger.warning(msg, user_id)
            raise AuthenticationFailedError(details=(msg % (user_id,)))

        session_id = jwt_payload.get("session_id")
        if not session_id:
            raise AuthenticationFailedError("Incorrect data in JWT: session_id is missed")

        query = select(UserSession).filter(
            UserSession.public_id == session_id,
            UserSession.is_active.is_(True),
        )
        user_session = (await self.request.state.db_session.execute(query)).scalars().first()
        if not user_session:
            raise AuthenticationFailedError(
                f"Couldn't found active session: {user_id=} | {session_id=}."
            )

        return user, jwt_payload, session_id


class SessionJWTAuthenticationBackend(JWTAuthenticationBackend):
    cookie_name = "session"

    async def authenticate(self, **kwargs) -> User:
        cookie_value = self.request.cookies.get(self.cookie_name)
        if not cookie_value:
            raise AuthenticationRequiredError("Cookie not found or is empty.")

        user, _, session_id = await self.authenticate_user(jwt_token=cookie_value, **kwargs)

        self.scope["user_session_id"] = session_id
        self.scope["user"] = user
        return user



from webargs_starlette import StarletteParser

from starlette_web.contrib.camel_case.utils import underscoreize


class CamelCaseStarletteParser(StarletteParser):
    async def _async_load_location_data(self, schema, req, location):
        data = await super()._async_load_location_data(schema, req, location)
        return underscoreize(data, no_underscore_before_number=True)



from importlib import import_module
from typing import Any


def import_string(dotted_path) -> Any:
    try:
        module_path, class_name = dotted_path.rsplit(".", 1)
    except ValueError as err:
        raise ImportError("%s doesn't look like a module path" % dotted_path) from err

    module = import_module(module_path)

    try:
        return getattr(module, class_name)
    except AttributeError as err:
        error_description = 'Module "%s" does not define a "%s" attribute/class' % (
            module_path,
            class_name,
        )
        raise ImportError(error_description) from err



# flake8: noqa

import argparse
import os


os.environ.setdefault("STARLETTE_SETTINGS_MODULE", "core.settings")
from starlette_web.common.app import get_asgi_application

app = get_asgi_application()


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--settings", required=False)
    args, _ = parser.parse_known_args()
    if args.settings:
        os.environ.setdefault("STARLETTE_SETTINGS_MODULE", args.settings)

    import uvicorn

    uvicorn.run(app, host="127.0.0.1", port=80)



import datetime
from dataclasses import dataclass
from functools import cached_property

from starlette_web.common.conf import settings
from starlette_web.common.utils.json import StarletteJSONEncoder
from starlette_web.contrib.auth.jwt_utils import JWTProcessor


TOKEN_TYPE_ACCESS = "access"
TOKEN_TYPE_REFRESH = "refresh"
TOKEN_TYPE_RESET_PASSWORD = "reset_password"


@dataclass
class TokenCollection:
    refresh_token: str
    refresh_token_expired_at: datetime.datetime
    access_token: str
    access_token_expired_at: datetime.datetime


class AuthJWTProcessor(JWTProcessor):
    @cached_property
    def _get_encode_secret_key(self):
        return str(settings.SECRET_KEY)

    @cached_property
    def _get_decode_secret_key(self):
        return str(settings.SECRET_KEY)

    def _get_expires_at(self, expires_in: int = None, **kwargs) -> datetime.datetime:
        token_type: str = kwargs.get("token_type", TOKEN_TYPE_ACCESS)

        if token_type == TOKEN_TYPE_REFRESH:
            expires_in = settings.AUTH_JWT_REFRESH_EXPIRES_IN
        else:
            expires_in = expires_in or settings.AUTH_JWT_EXPIRES_IN

        return super()._get_expires_at(expires_in=expires_in)

    def _enhance_payload_for_encode(self, payload: dict, **kwargs) -> None:
        token_type: str = kwargs.get("token_type", TOKEN_TYPE_ACCESS)
        payload["exp_iso"] = payload["exp"].isoformat()
        payload["token_type"] = token_type


jwt_processor = AuthJWTProcessor(
    algorithm=settings.AUTH_JWT_ALGORITHM,
    verify_signature=True,
    json_encoder=StarletteJSONEncoder,
)
encode_jwt = jwt_processor.encode_jwt
decode_jwt = jwt_processor.decode_jwt



import copy
import datetime
from functools import cached_property
from typing import Tuple

import jwt

from starlette_web.common.utils.inspect import get_available_options


class JWTProcessor:
    def __init__(self, **kwargs):
        self.init_options = kwargs

    def encode_jwt(
        self,
        payload: dict,
        expires_in: int = None,
        **kwargs,
    ) -> Tuple[str, datetime.datetime]:
        _expires_at = self._get_expires_at(
            expires_in=expires_in,
            **self.init_options,
            **kwargs,
        )
        _payload = copy.deepcopy(payload)
        _payload["exp"] = _expires_at
        self._enhance_payload_for_encode(
            _payload,
            **self.init_options,
            **kwargs,
        )

        token = jwt.encode(
            payload=_payload,
            key=self._get_encode_secret_key,
            **self._get_encode_options(**self.init_options, **kwargs),
        )
        return token, _expires_at

    def decode_jwt(self, encoded_jwt: str, **kwargs):
        return jwt.decode(
            encoded_jwt,
            key=self._get_decode_secret_key,
            **self._get_decode_options(**self.init_options, **kwargs),
        )

    @cached_property
    def _get_encode_secret_key(self):
        raise NotImplementedError()

    @cached_property
    def _get_decode_secret_key(self):
        raise NotImplementedError()

    def _get_expires_at(self, expires_in: int = None, **kwargs) -> datetime.datetime:
        return datetime.datetime.utcnow() + datetime.timedelta(seconds=expires_in)

    def _enhance_payload_for_encode(self, payload: dict, **kwargs) -> None:
        pass

    def _get_encode_options(self, **kwargs) -> dict:
        res = {}
        options = {**self.init_options, **kwargs}

        for _option in get_available_options(jwt.encode):
            if _option in options:
                res[_option] = options[_option]

        return res

    def _get_decode_options(self, **kwargs) -> dict:
        res = {}
        options = {**self.init_options, **kwargs}

        for _option in get_available_options(jwt.decode):
            if _option in options:
                res[_option] = options[_option]

        if "algorithm" in options and not res.get("algorithms"):
            res["algorithms"] = [options["algorithm"]]

        if "options" not in res:
            res["options"] = options

        return res



from starlette.requests import HTTPConnection
from starlette.types import Scope

from starlette_web.common.authorization.base_user import BaseUserMixin, AnonymousUser


class BaseAuthenticationBackend:
    openapi_spec = None
    openapi_name = "Base"

    def __init__(self, request: HTTPConnection, scope: Scope):
        self.request: HTTPConnection = request
        self.scope: Scope = scope

    async def authenticate(self, **kwargs) -> BaseUserMixin:
        raise NotImplementedError


class NoAuthenticationBackend(BaseAuthenticationBackend):
    openapi_name = "NoAuth"

    async def authenticate(self, **kwargs) -> BaseUserMixin:
        return AnonymousUser()



import os
import sys
from pathlib import Path
from typing import List, AnyStr, AsyncContextManager, Union

import anyio
from anyio._core._fileio import AsyncFile

from starlette_web.common.conf import settings
from starlette_web.common.files.storages.base import BaseStorage, MODE
from starlette_web.common.files.filelock import FileLock
from starlette_web.common.http.exceptions import ImproperlyConfigured
from starlette_web.common.utils import urljoin


class FilesystemStorage(BaseStorage):
    def __init__(self, **options):
        super().__init__(**options)
        self.BASE_DIR = self.options.get("BASE_DIR")
        self._initialize_base_dir()

    def _initialize_base_dir(self):
        if self.BASE_DIR is None:
            raise ImproperlyConfigured(details="Storage must be inited with BASE_DIR.")

        try:
            self.BASE_DIR = Path(self.BASE_DIR)
            self.BASE_DIR.mkdir(exist_ok=True, parents=True)
        except OSError as exc:
            raise ImproperlyConfigured(details=str(exc)) from exc

    def _normalize_path(self, path: Union[str, Path]) -> anyio.Path:
        return anyio.Path(str(self.BASE_DIR / str(Path(path)).strip(os.sep)))

    async def delete(self, path: str):
        async with self.get_access_lock(path, mode="w"):
            _path = self._normalize_path(path)
            if await _path.is_dir():
                await _path.rmdir()
            elif await _path.is_file():
                await _path.unlink()

    async def listdir(self, path: str) -> List[str]:
        async with self.get_access_lock(path, mode="r"):
            _path = self._normalize_path(path)
            _paths = []
            async for path in _path.iterdir():
                _paths.append(path.name)
        return _paths

    async def exists(self, path: str) -> bool:
        async with self.get_access_lock(path, mode="r"):
            _path = self._normalize_path(path)
            return await _path.exists()

    async def size(self, path: str) -> int:
        async with self.get_access_lock(path, mode="r"):
            _path = self._normalize_path(path)
            return (await _path.stat()).st_size

    async def get_mtime(self, path) -> float:
        async with self.get_access_lock(path):
            _path = self._normalize_path(path)
            return (await _path.stat()).st_mtime

    async def _open(self, path: str, mode: MODE = "b", **kwargs) -> AsyncFile:
        _path = self._normalize_path(path)
        await self._mkdir(os.path.dirname(_path), **kwargs)
        async_file: AsyncFile = await anyio.open_file(_path, mode, **kwargs)
        return await async_file.__aenter__()

    async def _close(self, fd: AsyncFile) -> None:
        await fd.__aexit__(*sys.exc_info())

    async def _write(self, fd: AsyncFile, content: AnyStr) -> None:
        await fd.write(content)

    async def _read(self, fd: AsyncFile, size: int = -1) -> AnyStr:
        return await fd.read(size)

    async def _readline(self, fd: AsyncFile, size: int = -1) -> AnyStr:
        # TODO: buffered read (anyio AsyncFile does not support size)
        return await fd.readline()

    async def _mkdir(self, path: str, **kwargs) -> None:
        _path = self._normalize_path(path)
        mode = kwargs.pop("mode", self.directory_create_mode)
        exist_ok = kwargs.pop("exist_ok", True)
        parents = kwargs.pop("parents", True)
        await _path.mkdir(exist_ok=exist_ok, parents=parents, mode=mode)

    async def _finalize_write(self, fd: AsyncFile) -> None:
        await fd.flush()
        os.fsync(fd._fp.fileno())

    def get_access_lock(self, path: str, mode="r") -> AsyncContextManager:
        # Consider subclassing FileSystemStorage, to use
        # faster cross-process lock, i.e. Redis lock
        if mode == "w":
            _path = Path(self._normalize_path(path))
            _lockname = _path.parent / (_path.name + ".lock")

            return FileLock(
                name=str(_lockname),
                timeout=self.write_timeout,
                blocking_timeout=self.blocking_timeout,
            )
        else:
            return super().get_access_lock(path, mode)


class MediaFileSystemStorage(FilesystemStorage):
    def __init__(self, **options):
        BaseStorage.__init__(self, **options)
        self.BASE_DIR = settings.MEDIA["ROOT_DIR"]
        self._initialize_base_dir()

    async def get_url(self, path: str) -> str:
        _path = str(Path(path)).split(os.sep)
        return urljoin(settings.MEDIA["URL"], "/".join(_path))



import os
import shutil
from pathlib import Path

from starlette_web.common.management.alembic_mixin import AlembicMixin
from starlette_web.common.management.base import BaseCommand, CommandError, CommandParser
from starlette_web.common.utils import get_random_string
from jinja2 import Template


class Command(BaseCommand, AlembicMixin):
    help = "Initialize directory with project files"
    _alembic_directory_name = "alembic"

    def add_arguments(self, parser: CommandParser):
        parser.add_argument("project_name", type=str)
        parser.add_argument(
            "--migration_prefix",
            type=str,
            default="auto",
            choices=["auto", "date"],
            help="Prefix for migration files' names.\n"
                 "'auto' - django-like ordered numerical prefix 0001, 0002, 0003, ...\n"
                 "'date' - datetime of file creation",
        )

    async def handle(self, **options):
        project_name = options["project_name"]
        migration_prefix = options.get("migration_prefix")

        project_dir = Path(os.getcwd()) / project_name
        if project_dir.is_file() or project_dir.is_symlink():
            raise CommandError(
                details=(
                    f"Cannot create project directory {project_name}. "
                    "A file/link with such name exists in the current directory."
                )
            )

        if project_dir.is_dir():
            raise CommandError(details=f"Directory {project_dir} already exists. Exiting.")

        project_dir.mkdir()
        defaults_dir = Path(__file__).parent / "_project_defaults"

        shutil.copytree(
            defaults_dir / "core",
            project_dir / "core",
        )
        for filename in ["command.py", "asgi.py", "__init__.py"]:
            shutil.copy(
                defaults_dir / filename,
                project_dir / filename,
            )

        # Setup base directories
        (project_dir / "static").mkdir()
        (project_dir / "templates").mkdir()

        # Setup env files
        env_template_content = self._read_template_file(file_name=defaults_dir / ".env_template")
        with open(project_dir / ".env", "wt+", encoding="utf-8") as file:
            file.write(env_template_content.format(secret_key=get_random_string(50)))

        with open(project_dir / ".env.template", "wt+", encoding="utf-8") as file:
            file.write(env_template_content.format(secret_key=""))

        # Setup alembic
        os.chdir(project_dir)
        await self.run_alembic_main(["init", "-t", "async", self._alembic_directory_name])
        await self._setup_alembic_conf(
            project_dir=project_dir,
            env_file_path=defaults_dir / "alembic" / "env.py-tpl",
            migration_prefix=migration_prefix,
        )

    @staticmethod
    def _read_template_file(file_name: Path) -> str:
        if not (file_name.exists() and file_name.is_file()):
            raise CommandError(details=f"Invalid file template path: {str(file_name)}")

        with file_name.open() as file:
            return file.read()

    async def _setup_alembic_conf(
        self,
        project_dir: Path,
        env_file_path: Path,
        migration_prefix: str,
    ) -> None:
        alembic_env_content = Template(self._read_template_file(file_name=env_file_path))
        with open(project_dir / self._alembic_directory_name / "env.py", "wt") as file:
            file.write(alembic_env_content.render(migration_prefix=migration_prefix))

        with open(project_dir / "alembic.ini", "rt") as file:
            lines = []
            for line in file:
                if migration_prefix == "date" and "# file_template = " in line:
                    lines.append(line[2:])
                elif migration_prefix == "auto" and "# revision_environment = false" in line:
                    lines += "revision_environment = true"
                else:
                    lines.append(line)

        with open(project_dir / "alembic.ini", "wt") as file:
            file.writelines(lines)



from email.mime.base import MIMEBase
from typing import Optional, Sequence

import anyio

from starlette_web.common.http.exceptions import BaseApplicationError


class EmailSenderError(BaseApplicationError):
    status_code = 503
    message = "Error when sending email."


class BaseEmailSender:
    MAX_BULK_SIZE = 1
    EXIT_MAX_DELAY = 60

    def __init__(self, **options):
        self._connection_opened = False
        self.options = options

    async def _open(self):
        return self

    async def _close(self):
        pass

    async def __aenter__(self):
        try:
            if not self._connection_opened:
                await self._open()
                self._connection_opened = True
            return self

        except Exception as exc:
            with anyio.fail_after(self.EXIT_MAX_DELAY, shield=True):
                if self._connection_opened:
                    await self._close()
                    self._connection_opened = False

            raise EmailSenderError(details=str(exc)) from exc

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        with anyio.fail_after(self.EXIT_MAX_DELAY, shield=True):
            if self._connection_opened:
                await self._close()
                self._connection_opened = False

        if exc_type:
            raise EmailSenderError(details=str(exc_val)) from exc_val

    async def send_email(
        self,
        subject: str,
        html_content: str,
        recipients_list: Sequence[str],
        from_email: Optional[str] = None,
        attachments: Optional[Sequence[MIMEBase]] = None,
        **kwargs,
    ):
        raise NotImplementedError



from starlette_web.common.conf import settings
from starlette_web.common.conf.base_app_config import BaseAppConfig
from starlette_web.common.http.exceptions import BaseApplicationError, ImproperlyConfigured


class AppConfig(BaseAppConfig):
    app_name = "constance"
    _error_message = (
        "settings.CONSTANCE_CONFIG must be of type " "Dict[str, Tuple[Any, str, Type]]."
    )

    def perform_checks(self):
        try:
            config = settings.CONSTANCE_CONFIG
        except (AttributeError, BaseApplicationError):
            config = {}

        self._run_config_checks(config)

    def _run_config_checks(self, _config):
        try:
            assert type(_config) is dict

            for key, value in _config.items():
                assert type(key) is str
                actual_value, description, value_type = value
                assert isinstance(value_type, type)
                assert isinstance(actual_value, value_type)
                assert type(description) is str
        except (AssertionError, TypeError, ValueError):
            raise ImproperlyConfigured(details=self._error_message)



# Adapted from https://github.com/sabuhish/fastapi-mqtt/blob/master/fastapi_mqtt/fastmqtt.py

import logging
import ssl
import uuid
from itertools import zip_longest
from typing import Any, Optional, Union, Literal, Type, Set

import anyio
from anyio.streams.memory import (
    MemoryObjectReceiveStream,
    MemoryObjectSendStream,
    ClosedResourceError,
)
from gmqtt import Client as MQTTClient
from gmqtt.mqtt.connection import MQTTConnection
from gmqtt.mqtt.constants import MQTTv50, MQTTv311
from gmqtt.mqtt.protocol import MQTTProtocol
from gmqtt.mqtt.property import PROPERTIES_BY_NAME

from starlette_web.common.channels.event import Event
from starlette_web.common.channels.exceptions import ListenerClosed
from starlette_web.common.channels.layers.base import BaseChannelLayer
from starlette_web.common.utils.inspect import get_available_options
from starlette_web.common.utils.serializers import BytesSerializer
from starlette_web.contrib.mqtt.serializers import MQTTSerializer


logger = logging.getLogger("starlette_web.contrib.mqtt")


class MQTTChannelLayer(BaseChannelLayer):
    """
    Experimental implementation of MQTT support.

    Underlying library may be changed
    (to anyio client, when such is developed and out of pre-alpha,
    or to aiomqtt, once issue with Windows.ProactorEventLoop is fixed)

    NOTE: gmqtt has problems with auto-switching to 4-th version of protocol,
    so you have to manually set it.
    """

    client: MQTTClient
    _available_options = list(
        {
            "retry_deliver_timeout",
            "persistent_storage",
            "topic_alias_maximum",
            *get_available_options(MQTTConnection.auth),
            *get_available_options(MQTTProtocol.send_auth_package),
            *list(PROPERTIES_BY_NAME.keys()),
        }
    )
    serializer_class: Type[BytesSerializer] = MQTTSerializer
    _connection_closed_flag = object()

    def __init__(self, **options):
        super().__init__(**options)

        self.host: str = options.pop("host", "localhost")
        self.port: int = options.pop("port", 1883)
        self.ssl: Union[bool, ssl.SSLContext] = options.pop("ssl", False)
        self.keepalive: int = options.pop("keepalive", 60)
        self.username: Optional[str] = options.pop("username", None)
        self.password: Optional[str] = options.pop("password", None)
        self.version: Literal[MQTTv50, MQTTv311] = options.pop("version", MQTTv50)

        self.reconnect_retries: Optional[int] = options.pop("reconnect_retries", 1)
        self.reconnect_delay: Optional[int] = options.pop("reconnect_delay", 6)

        self.client_id = options.get("client_id", uuid.uuid4().hex)
        self.client = MQTTClient(
            self.client_id,
            **{key: value for key, value in options.items() if key in self._available_options},
        )

        self.clean_session: bool = options.get("clean_session", True)
        self.optimistic_acknowledgement: bool = options.get("optimistic_acknowledgement", True)

        self.client._clean_session = self.clean_session
        self.client._username = self.username
        self.client._password = self.password
        self.client._host = self.host
        self.client._port = self.port
        self.client._keepalive = self.keepalive
        self.client._ssl = self.ssl
        self.client.optimistic_acknowledgement = self.optimistic_acknowledgement
        self.subscriptions: Set[str] = set()

        self.client.on_connect = self._on_connect
        self.client.on_message = self._on_message
        self.client.on_disconnect = self._on_disconnect
        self.client.on_subscribe = self._on_subscribe

        self._receive_stream: Optional[MemoryObjectReceiveStream] = None
        self._send_stream: Optional[MemoryObjectSendStream] = None
        self._serializer = self.serializer_class()

    def __str__(self):
        return f"{MQTTChannelLayer} {self.host}:{self.port} [{self.client_id}]"

    async def connect(self) -> None:
        self._send_stream, self._receive_stream = anyio.create_memory_object_stream()

        if self.client._username:
            self.client.set_auth_credentials(self.client._username, self.client._password)
            logger.debug("User is authenticated")

        self.client.set_config(
            {
                "reconnect_retries": self.reconnect_retries,
                "reconnect_delay": self.reconnect_delay,
            }
        )

        version = self.version or MQTTv50
        logger.info("Used broker version is %s", version)

        await self.client.connect(
            self.client._host,
            self.client._port,
            self.client._ssl,
            self.client._keepalive,
            version,
        )
        logger.debug(f"Connected to {self}")

    async def disconnect(self) -> None:
        try:
            await self.client.disconnect()
        finally:
            self._receive_stream.close()
            self._send_stream.close()
            self._receive_stream = None
            self._send_stream = None

    async def subscribe(
        self,
        topic: str,
        qos: int = 0,
        no_local: bool = False,
        retain_as_published: bool = False,
        retain_handling_options: int = 0,
        subscription_identifier: Any = None,
        **kwargs,
    ) -> None:
        logger.debug(f"Subscribe to {topic}")
        self.subscriptions.add(topic)
        self.client.subscribe(topic)

    async def unsubscribe(self, topic: str, **kwargs) -> None:
        """
        Defined to unsubscribe topic

        topic: topic name
        """
        logger.debug(f"Unsubscribe from {topic}")
        if topic in self.subscriptions:
            self.subscriptions.discard(topic)

        return self.client.unsubscribe(topic, **kwargs)

    async def publish(
        self,
        topic: str,
        message: Any,
        qos: int = 0,
        retain: bool = False,
        **kwargs,
    ) -> None:
        """
        Defined to publish payload MQTT server

        message_or_topic: topic name

        payload: message payload

        qos: Quality of Assurance

        retain:
        """
        _message = self._serializer.serialize(message)
        return self.client.publish(topic, payload=_message, qos=qos, retain=retain, **kwargs)

    async def next_published(self) -> Event:
        try:
            event: Event = await self._receive_stream.receive()
        except ClosedResourceError as exc:
            raise ListenerClosed from exc

        return event

    def _on_subscribe(self, client, mid, qos, properties):
        logger.debug("SUBSCRIBED")

    def _on_connect(self, client: MQTTClient, flags: int, rc: int, properties: Any):
        """
        Generic on connecting handler, it would call user handler if defined.
        Will perform subscription for given topics.
        It cannot be done earlier, since subscription relies on connection.
        """
        for topic in self.subscriptions:
            logger.debug(f"Subscribing for {topic}")
            self.client.subscribe(topic)

    async def _on_message(
        self,
        client: MQTTClient,
        topic: str,
        payload: bytes,
        qos: int,
        properties: Any,
    ):
        _decoded = self._serializer.deserialize(payload)
        logger.debug(f"Received message: {topic}, {_decoded}, {qos}, {properties}")

        for template in self.subscriptions.copy():
            if self.match(topic, template):
                event = Event(group=template, message=_decoded)
                await self._send_stream.send(event)

    def _on_disconnect(self, client, packet, exc=None):
        logger.debug(f"Disconnected from {self}")
        self._send_stream.close()

    @staticmethod
    def match(topic: str, template: str) -> bool:
        """
        Defined match topics

        topic: topic name
        template: template topic name that contains wildcards
        """
        if str(template).startswith("$share/"):
            template = template.split("/", 2)[2]

        topic_parts = topic.split("/")
        template_parts = template.split("/")

        for topic_part, part in zip_longest(topic_parts, template_parts):
            if part == "#" and not str(topic_part).startswith("$"):
                return True
            elif (topic_part is None or part not in {"+", topic_part}) or (
                part == "+" and topic_part.startswith("$")
            ):
                return False
            continue

        return len(template_parts) == len(topic_parts)



from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from typing import Optional, Sequence

import aiosmtplib
import anyio

from starlette_web.common.email.base_sender import BaseEmailSender
from starlette_web.common.http.exceptions import NotSupportedError
from starlette_web.common.utils import safe_init


class SMTPEmailSender(BaseEmailSender):
    SEND_MAX_TIME = 30
    _client: Optional[aiosmtplib.SMTP] = None

    async def _open(self):
        self._client = safe_init(aiosmtplib.SMTP, **self.options)
        await self._client.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self._client.__aexit__(exc_type, exc_val, exc_tb)

    async def send_email(
        self,
        subject: str,
        html_content: str,
        recipients_list: Sequence[str],
        from_email: Optional[str] = None,
        attachments: Optional[Sequence[MIMEBase]] = None,
        **kwargs,
    ):
        message = MIMEMultipart("alternative")
        message["Subject"] = subject
        _from = from_email
        if _from is None:
            _from = self.options.get("from")
        if _from is None:
            raise NotSupportedError(details="Cannot send email without setting FROM field.")
        message["From"] = _from

        html_part = MIMEText(html_content, "html")
        message.attach(html_part)

        if attachments:
            for attachment in attachments:
                message.attach(attachment)

        _recipients = list(recipients_list).copy()

        while _recipients:
            with anyio.move_on_after(self.SEND_MAX_TIME):
                await self._client.send_message(
                    message,
                    sender=str(_from),
                    recipients=recipients_list,
                    mail_options=kwargs.get("mail_options"),
                    rcpt_options=kwargs.get("rcpt_options"),
                    timeout=kwargs.get("timeout", aiosmtplib.typing._default),
                )
            del _recipients[: self.MAX_BULK_SIZE]



from typing import Optional, Any

from sqlalchemy.exc import IntegrityError
from starlette.applications import Starlette
from starlette.datastructures import URLPath
from starlette.exceptions import HTTPException
from starlette.middleware import Middleware
from starlette.requests import Request
from starlette.responses import Response
from starlette.routing import Mount
from starlette_admin.base import BaseAdmin
from starlette_admin.views import CustomView
from starlette_admin.exceptions import StarletteAdminException

from starlette_web.common.conf import settings
from starlette_web.common.utils import urljoin
from starlette_web.contrib.admin.auth_provider import AdminAuthProvider
from starlette_web.contrib.admin.middleware import (
    DBSessionMiddleware,
    AdminSessionMiddleware,
)


class AdminMount(Mount):
    def __init__(self, path, **kwargs):
        super().__init__(path, **kwargs)
        self._base_app.state.ROUTE_NAME = self.name

    # starlette_admin is not a built-in module of starlette_web,
    # so it requires specific workaround for static files
    def url_path_for(self, name: str, **path_params: Any) -> URLPath:
        if (name == self.name + ":statics") and ("path" in path_params):
            path = urljoin(settings.STATIC["URL"], "admin", path_params["path"])
            return URLPath(path=path)
        return super().url_path_for(name, **path_params)


class Admin(BaseAdmin):
    def __init__(
        self,
        title: str = "Admin",
        logo_url: Optional[str] = None,
        login_logo_url: Optional[str] = None,
        index_view: Optional[CustomView] = None,
    ) -> None:
        super().__init__(
            title=title,
            base_url="/admin",
            route_name="admin",
            logo_url=logo_url,
            login_logo_url=login_logo_url,
            templates_dir=settings.TEMPLATES["ROOT_DIR"],
            statics_dir=None,
            index_view=index_view,
            auth_provider=AdminAuthProvider(),
            middlewares=[],
            debug=settings.APP_DEBUG,
        )
        self.middlewares = [] if self.middlewares is None else list(self.middlewares)
        # TODO: support CSRF protection
        self.middlewares = [
            Middleware(DBSessionMiddleware),
            Middleware(AdminSessionMiddleware),
        ] + self.middlewares

        # Remove static files from starlette_admin routing,
        # it is handled differently
        self.routes = [route for route in self.routes if type(route) is not Mount]

    def get_app(self) -> Starlette:
        for route in self.routes:
            if hasattr(route, "include_in_schema"):
                route.include_in_schema = False

        admin_app = Starlette(
            routes=self.routes,
            middleware=self.middlewares,
            debug=self.debug,
            exception_handlers={
                HTTPException: self._render_error,
                StarletteAdminException: self._render_error,
                IntegrityError: self._render_error,
            },
        )
        return admin_app

    async def _render_error(
        self,
        request: Request,
        exc: Exception = HTTPException(status_code=500),  # noqa: B008
    ) -> Response:
        try:
            _ = exc.status_code
            _ = exc.detail
        except AttributeError:
            exc.status_code = 500
            exc.detail = str(exc)

        return self.templates.TemplateResponse(
            request=request,
            name="error.html",
            context={"request": request, "exc": exc},
            status_code=exc.status_code,
        )



# Copied from https://github.com/encode/django-rest-framework/blob/master/rest_framework/permissions.py  # noqa E501

from typing import Type, Union

from starlette.requests import Request
from starlette.types import Scope

from starlette_web.common.authorization.base_user import BaseUserMixin


class OperationHolderMixin:
    def __and__(self, other):
        return OperandHolder(AND, self, other)

    def __or__(self, other):
        return OperandHolder(OR, self, other)

    def __rand__(self, other):
        return OperandHolder(AND, other, self)

    def __ror__(self, other):
        return OperandHolder(OR, other, self)

    def __invert__(self):
        return SingleOperandHolder(NOT, self)


class SingleOperandHolder(OperationHolderMixin):
    def __init__(self, operator_class, op1_class):
        self.operator_class = operator_class
        self.op1_class = op1_class

    def __call__(self, *args, **kwargs):
        op1 = self.op1_class(*args, **kwargs)
        return self.operator_class(op1)


class OperandHolder(OperationHolderMixin):
    def __init__(self, operator_class, op1_class, op2_class):
        self.operator_class = operator_class
        self.op1_class = op1_class
        self.op2_class = op2_class

    def __call__(self, *args, **kwargs):
        op1 = self.op1_class(*args, **kwargs)
        op2 = self.op2_class(*args, **kwargs)
        return self.operator_class(op1, op2)


class AND:
    def __init__(self, op1, op2):
        self.op1 = op1
        self.op2 = op2

    async def has_permission(self, request: Request, scope: Scope):
        return (await self.op1.has_permission(request, scope)) and (
            await self.op2.has_permission(request, scope)
        )


class OR:
    def __init__(self, op1, op2):
        self.op1 = op1
        self.op2 = op2

    async def has_permission(self, request: Request, scope: Scope):
        return (await self.op1.has_permission(request, scope)) or (
            await self.op2.has_permission(request, scope)
        )


class NOT:
    def __init__(self, op1):
        self.op1 = op1

    async def has_permission(self, request: Request, scope: Scope):
        return not (await self.op1.has_permission(request, scope))


class BasePermissionMetaclass(OperationHolderMixin, type):
    pass


class BasePermission(metaclass=BasePermissionMetaclass):
    async def has_permission(self, request: Request, scope: Scope) -> bool:
        raise NotImplementedError


class AllowAnyPermission(BasePermission):
    async def has_permission(self, request: Request, scope: Scope) -> bool:
        return True


class IsAuthenticatedPermission(BasePermission):
    async def has_permission(self, request: Request, scope: Scope) -> bool:
        if "user" not in scope:
            return False

        user: BaseUserMixin = scope["user"]
        return user.is_authenticated


PermissionType = Union[Type[BasePermission], OperandHolder]



def urljoin(first: str, *urls) -> str:
    res = first
    for url in urls:
        if "//" in url:
            res = url
        else:
            res = res.rstrip("/") + "/" + url.lstrip("/")
    return res



from typing import Optional, Any

import anyio
import asyncpg
from anyio.streams.memory import (
    MemoryObjectReceiveStream,
    MemoryObjectSendStream,
)

from starlette_web.common.channels.event import Event
from starlette_web.common.channels.exceptions import ListenerClosed
from starlette_web.common.channels.layers.base import BaseChannelLayer
from starlette_web.common.http.exceptions import ImproperlyConfigured, NotSupportedError
from starlette_web.common.utils import get_available_options


class PostgreSQLChannelLayer(BaseChannelLayer):
    MAX_MESSAGE_BYTELEN = 8000
    _connection_closed_flag = object()

    def __init__(self, **options) -> None:
        # Options must be valid kwargs for asyncpg.connect
        super().__init__(**options)
        self._conn: Optional[asyncpg.connection.Connection] = None
        self._manager_lock = anyio.Lock()
        self._receive_stream: Optional[MemoryObjectReceiveStream] = None
        self._send_stream: Optional[MemoryObjectSendStream] = None

        self._connection_options = {
            key: value
            for key, value in options.items()
            if key in get_available_options(asyncpg.connect)
        }

    async def connect(self) -> None:
        try:
            async with self._manager_lock:
                self._send_stream, self._receive_stream = anyio.create_memory_object_stream()
                self._conn = await asyncpg.connect(**self._connection_options)
                self._conn.add_termination_listener(self._termination_listener)
        except ValueError as exc:
            raise ImproperlyConfigured(details=str(exc)) from exc

    async def disconnect(self) -> None:
        async with self._manager_lock:
            try:
                await self._conn.close()
            finally:
                self._send_stream.close()
                self._receive_stream.close()
                self._send_stream = None
                self._receive_stream = None

    async def subscribe(self, group: str, **kwargs) -> None:
        await self._conn.add_listener(group, self._listener)

    async def _listener(self, *args: Any) -> None:
        connection, pid, channel, payload = args
        event = Event(group=channel, message=payload)
        await self._send_stream.send(event)

    async def unsubscribe(self, group: str, **kwargs) -> None:
        await self._conn.remove_listener(group, self._listener)

    async def publish(self, group: str, message: Any, **kwargs) -> None:
        await self._validate_message(message)
        await self._conn.execute("SELECT pg_notify($1, $2);", group, message)

    async def _termination_listener(self, *args) -> None:
        await self._send_stream.send(self._connection_closed_flag)

    async def _validate_message(self, message):
        # https://www.postgresql.org/docs/current/sql-notify.html
        if type(message) is not str:
            raise NotSupportedError(
                details="Publish message for PostgreSQL NOTIFY must be a string"
            )

        if len(message.encode("utf-8")) > self.MAX_MESSAGE_BYTELEN:
            raise NotSupportedError(
                details=f"Message byte-length must be at most "
                f"{self.MAX_MESSAGE_BYTELEN} symbols"
            )

    async def next_published(self) -> Event:
        event: Event = await self._receive_stream.receive()
        if event is self._connection_closed_flag:
            raise ListenerClosed
        return event



from starlette.routing import Route, Mount
from starlette_web.contrib.auth import views

routes = [
    Mount(
        "/auth",
        routes=[
            Route("/me/", views.ProfileApiView),
            Route("/sign-in/", views.SignInAPIView),
            Route("/sign-up/", views.SignUpAPIView),
            Route("/sign-out/", views.SignOutAPIView),
            Route("/refresh-token/", views.RefreshTokenAPIView),
            Route("/invite-user/", views.InviteUserAPIView),
            Route("/reset-password/", views.ResetPasswordAPIView),
            Route("/change-password/", views.ChangePasswordAPIView),
        ],
    )
]



from urllib.parse import urlencode

from sqlalchemy import select, update
from starlette.middleware import Middleware
from starlette.middleware.base import RequestResponseEndpoint
from starlette.requests import Request
from starlette.status import HTTP_303_SEE_OTHER
from starlette.responses import RedirectResponse, Response
from starlette_admin.auth import AdminUser, AuthProvider, AuthMiddleware
from starlette_admin.exceptions import LoginFailed

from starlette_web.common.http.exceptions import AuthenticationFailedError, PermissionDeniedError
from starlette_web.contrib.auth.backend import SessionJWTAuthenticationBackend
from starlette_web.contrib.auth.views import JWTSessionMixin
from starlette_web.contrib.auth.models import User, UserSession
from starlette_web.contrib.auth.permissions import IsSuperuserPermission
from starlette_web.contrib.auth.utils import TOKEN_TYPE_REFRESH, decode_jwt


class AdminAuthProvider(AuthProvider):
    allow_paths = []

    async def login(
        self,
        username: str,
        password: str,
        remember_me: bool,
        request: Request,
        response: Response,
    ) -> Response:
        query = select(User).filter(User.email == username, User.is_active.is_(True))
        user = (await request.state.session.execute(query)).scalars().first()
        if not user:
            raise LoginFailed("Invalid username or password")

        if not user.verify_password(password):
            raise LoginFailed("Invalid username or password")

        if not user.is_superuser:
            raise LoginFailed(PermissionDeniedError.message)

        token_maker = JWTSessionMixin()
        token_maker.db_session = request.state.session
        token_collection = await token_maker._create_session(user)

        query = (
            update(UserSession)
            .filter(UserSession.refresh_token == token_collection.refresh_token)
            .values(is_persistent=remember_me)
        )

        await request.state.session.execute(query)
        await request.state.session.commit()

        request.scope["session"] = {
            "token": token_collection.refresh_token,
            **decode_jwt(token_collection.refresh_token),
        }
        return response

    async def is_authenticated(self, request) -> bool:
        try:
            await SessionJWTAuthenticationBackend(
                request,
                request.scope,
            ).authenticate(
                token_type=TOKEN_TYPE_REFRESH,
            )
            return await IsSuperuserPermission().has_permission(request, request.scope)
        except (AuthenticationFailedError, PermissionDeniedError):
            return False

    def get_admin_user(self, request: Request) -> AdminUser:
        if "user" in request.scope:
            username = request.scope["user"].email
        else:
            username = "Anonymous"
        return AdminUser(username=username, photo_url=None)

    async def logout(self, request: Request, response: Response) -> Response:
        request.scope["session"] = {}
        return response

    def get_middleware(self, admin) -> Middleware:
        return Middleware(StarletteWebAuthMiddleware, provider=self)


class StarletteWebAuthMiddleware(AuthMiddleware):
    async def dispatch(
            self, request: Request, call_next: RequestResponseEndpoint
    ) -> Response:
        _path = request.scope["path"]
        if "?" in _path:
            _path = _path[:_path.find("?")]

        if all([
            _path not in self.allow_paths,
            not _path.rstrip("/").endswith("/login"),
        ]) and not (
                await self.provider.is_authenticated(request)
        ):
            return RedirectResponse(
                "{url}?{query_params}".format(
                    url=request.url_for(request.app.state.ROUTE_NAME + ":login"),
                    query_params=urlencode({"next": str(request.url)}),
                ),
                status_code=HTTP_303_SEE_OTHER,
            )
        return await call_next(request)



from typing import Optional

import httpx

from starlette_web.common.i18n import gettext as _


class BaseApplicationError(Exception):
    message: str = httpx.codes.INTERNAL_SERVER_ERROR.name
    details: Optional[str] = None
    status_code: int = 500

    def __init__(
        self,
        details: str = None,
        message: str = None,
        status_code: int = None,
    ):
        self.message = message or self.message
        self.details = details or self.details
        self.status_code = status_code or self.status_code

    def __str__(self):
        return f"{self.message}\n{self.details}".strip()

    def __iter__(self):
        yield "message", self.message
        yield "details", self.details
        yield "status_code", self.status_code


class ImproperlyConfigured(BaseApplicationError):
    message = _("Application is configured improperly.")


class UnexpectedError(BaseApplicationError):
    message = _("Something unexpected happened.")


class NotSupportedError(BaseApplicationError):
    message = _("Requested action is not supported now.")


class InvalidParameterError(BaseApplicationError):
    status_code = 400
    message = _("Requested data is not valid.")


class AuthenticationFailedError(BaseApplicationError):
    status_code = 401
    message = _("Authentication credentials are invalid.")


class AuthenticationRequiredError(AuthenticationFailedError):
    message = _("Authentication is required.")


class SignatureExpiredError(AuthenticationFailedError):
    message = _("Authentication credentials have expired.")


class InviteTokenInvalidationError(AuthenticationFailedError):
    message = _("Requested token is expired or does not exist.")


class PermissionDeniedError(BaseApplicationError):
    status_code = 403
    message = _("You do not have permission to perform this action.")


class NotFoundError(BaseApplicationError):
    status_code = 404
    message = _("Requested object not found.")


class MethodNotAllowedError(BaseApplicationError):
    status_code = 405
    message = _("Requested method is not allowed.")


class NotAcceptableError(BaseApplicationError):
    status_code = 406
    message = _("Request cannot be processed, " "Accept-* headers are incompatible with server.")


class ConflictError(BaseApplicationError):
    status_code = 409
    message = _("Request conflicts with current state of server.")


class UnprocessableEntityError(BaseApplicationError):
    status_code = 422
    message = _("Could not process request due to logical errors in data.")


class InvalidResponseError(BaseApplicationError):
    status_code = 500
    message = _("Response data could not be serialized.")


class NotImplementedByServerError(BaseApplicationError):
    status_code = 501
    message = _("Functionality is not supported by the server.")


class HttpError(BaseApplicationError):
    status_code = 502
    message = _("Some HTTP error happened.")


class SendRequestError(BaseApplicationError):
    status_code = 503
    message = _("Got unexpected error for sending request.")


class MaxAttemptsReached(BaseApplicationError):
    status_code = 503
    message = _("Reached max attempt to make action")



from typing import AnyStr, List

from starlette_web.common.conf import settings
from starlette_web.common.files.storages.base import BaseStorage
from starlette_web.common.utils import import_string


class _StorageManager:
    def __init__(self):
        self._storages = {}

    def _get_storage(self, storage_name: str) -> BaseStorage:
        if storage_name in self._storages:
            return self._storages[storage_name]

        _conf = settings.STORAGES[storage_name]
        self._storages[storage_name] = import_string(_conf["BACKEND"])(
            **_conf.get("OPTIONS", {}),
        )
        return self._storages[storage_name]

    async def read(self, path: str, using="default", **kwargs) -> AnyStr:
        async with self._get_storage(using) as storage:
            async with storage.reader(path=path, **kwargs) as reader:
                return await reader.read()

    async def write(self, path: str, content: AnyStr = b"", using="default", **kwargs):
        async with self._get_storage(using) as storage:
            async with storage.writer(path=path, append=False, **kwargs) as writer:
                return await writer.write(content)

    async def append(self, path: str, content: AnyStr = b"", using="default", **kwargs):
        async with self._get_storage(using) as storage:
            async with storage.writer(path=path, append=True, **kwargs) as writer:
                return await writer.write(content)

    async def delete(self, path: str, using="default"):
        async with self._get_storage(using) as storage:
            return await storage.delete(path)

    async def get_url(self, path: str, using="default") -> str:
        async with self._get_storage(using) as storage:
            return await storage.get_url(path)

    async def listdir(self, path: str, using="default") -> List[str]:
        async with self._get_storage(using) as storage:
            return await storage.listdir(path)

    async def exists(self, path: str, using="default") -> bool:
        async with self._get_storage(using) as storage:
            return await storage.exists(path)

    async def size(self, path: str, using="default") -> int:
        async with self._get_storage(using) as storage:
            return await storage.size(path)

    async def get_mtime(self, path: str, using="default") -> float:
        async with self._get_storage(using) as storage:
            return await storage.get_mtime(path)


storage_manager = _StorageManager()



# flake8: noqa

from starlette_web.common.database import ModelBase


"""
Initialize your models in this file.
It will be automatically introspected, 
if you add %app_name% from "apps.py" to settings.INSTALLED_APPS

>>> class Model(ModelBase):
>>>     __tablename__ = 'app_model'
>>>     ...
"""
